#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
import symengine as se


class DiagonalOp(mc.AbstractOperator):
    r"""One-qubit diagonal operator.

    The corresponding matrix is given by:

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            a & 0 \\
            0 & b
        \end{pmatrix}

    where `a` and `b` are complex numbers.

    See Also:
        :class:`Operator`, :class:`Projector0`, :class:`Projector1`

    Parameters:
        a (complex): The top-left entry of the diagonal matrix.
        b (complex): The bottom-right entry of the diagonal matrix.

    Examples:
        >>> from mimiqcircuits import *
        >>> op = DiagonalOp(1, 0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(DiagonalOp(1, 0.5)), 1, 2)
        2-qubit circuit with 1 instructions:
        └── ⟨D(1, 0.5)⟩ @ q[1], z[2]
        <BLANKLINE>
    """

    _name = "D"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a, b):
        if not isinstance(a, (float, int)) or not isinstance(b, (float, int)):
            raise ValueError("a and b must be float or int")
        self.a = a
        self.b = b

        super().__init__()
        self._parnames = ("a", "b")

    def _matrix(self):
        mat = se.Matrix([[self.a, 0], [0, self.b]])
        return mat

    def __str__(self):
        return f"{self._name}{self.a,self.b}"

    @property
    def parnames(self):
        return self._parnames

    def rescale(self, scale):
        """Return a new rescaled DiagonalOp operator."""
        return DiagonalOp(self.a * scale, self.b * scale)

    def rescale_inplace(self, scale):
        """In-place rescaling of the DiagonalOp operator."""
        self.a *= scale
        self.b *= scale
        return self

    def opsquared(self):
        """Return the operator with each parameter squared."""
        return DiagonalOp(abs(self.a)**2, abs(self.b)**2)



__all__ = ["DiagonalOp"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.utils import control_one_defined
import mimiqcircuits as mc
from symengine import pi


class GateSY(mc.Power):
    r"""Single qubit :math:`\sqrt{Y}` gate.

    See Also:
        :class:`GateSYDG`, :class:`GateY`, :class:`Power`

    **Matrix Representation**

    .. math::
        \operatorname{SY} =
        \sqrt{\operatorname{Y}} =
        \frac{1}{2}
        \begin{pmatrix}
            1+i & -1-i \\
            1+i & 1+i
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateSY()
        SY

        >>> GateSY().matrix()
        [0.5 + 0.5*I, -0.5 - 0.5*I]
        [0.5 + 0.5*I, 0.5 + 0.5*I]
        <BLANKLINE>

        >>> c = Circuit()
        >>> c.push(GateSY(), 1)
        2-qubit circuit with 1 instructions:
        └── SY @ q[1]
        <BLANKLINE>

        >>> c.push(GateSY(), 2)
        3-qubit circuit with 2 instructions:
        ├── SY @ q[1]
        └── SY @ q[2]
        <BLANKLINE>

        >>> power(GateSY(), 2)
        Y**1.0

    **Decomposition**

        >>> GateSY().decompose()
        1-qubit circuit with 4 instructions:
        ├── S @ q[0]
        ├── S @ q[0]
        ├── H @ q[0]
        └── U(0, 0, 0, (1/4)*pi) @ q[0]
        <BLANKLINE>
      
    """

    _name = "SY"

    def __init__(self):
        super().__init__(mc.GateY(), 1 / 2)

    def inverse(self):
        return GateSYDG()

    def isopalias(self):
        return True

    def _control(self, n):
        return control_one_defined(n, self, mc.Control(1, GateSY()))

    def _power(self, p):
        return mc.Power(self, p)

    def __str__(self):
        return f"{self.name}"

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateS(), q)
        circ.push(mc.GateS(), q)
        circ.push(mc.GateH(), q)
        circ.push(mc.GateU(0, 0, 0, pi / 4), q)
        return circ


class GateSYDG(mc.Inverse):
    r"""Single qubit :math:`\sqrt{Y}^\dagger` gate (conjugate transpose of the :math:`\sqrt{Y}` gate).

    See Also:
        :class:`GateSY`, :class:`GateY`, :class:`Power`, :class:`Inverse`

    **Matrix Representation**

    .. math::
        \operatorname{SYDG} =
        \sqrt{\operatorname{Y}}^\dagger =
        \frac{1}{2}
        \begin{pmatrix}
            1-i & 1-i \\
            -1+i & 1-i
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateSYDG()
        SY†

        >>> GateSYDG().matrix()
        [0.5 - 0.5*I, 0.5 - 0.5*I]
        [-0.5 + 0.5*I, 0.5 - 0.5*I]
        <BLANKLINE>

        >>> c = Circuit()
        >>> c.push(GateSYDG(), 1)
        2-qubit circuit with 1 instructions:
        └── SY† @ q[1]
        <BLANKLINE>

        >>> c.push(GateSYDG(), 2)
        3-qubit circuit with 2 instructions:
        ├── SY† @ q[1]
        └── SY† @ q[2]
        <BLANKLINE>

        >>> power(GateSYDG(), 2)
        SY†**2

        >>> inverse(GateSYDG())
        SY

    """

    def __init__(self):
        super().__init__(GateSY())

    def inverse(self):
        return GateSY()

    def isopalias(self):
        return True

    def _power(self, p):
        return mc.Power(self, p)

    def _control(self, n):
        return control_one_defined(n, self, mc.Control(1, GateSYDG()))

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateU(0, 0, 0, -pi / 4), q)
        circ.push(mc.GateH(), q)
        circ.push(mc.GateSDG(), q)
        circ.push(mc.GateSDG(), q)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
from mimiqcircuits.lazy import LazyExpr, LazyArg


class AbstractAnnotation(mc.Operation):
    r"""AbstractAnnotation
    ----------------------

    Abstract base class for annotations in quantum circuits.

    This class is used as a base for defining various annotation types, such as `Detector`, `QubitCoordinates`, 
    `ShiftCoordinates`,`Tick`, `ObservableInclude`, which provide metadata or structural information for quantum circuits.
    """
    pass


class Detector(AbstractAnnotation):
    r"""Detector operation.
    ------------------------

    An annotation class representing a detector in a quantum circuit.

    The `Detector` operation monitors the parity of measurement results over `N` classical bits, where the parity should remain deterministic under ideal, noiseless execution. 
    The Detector ensures that the combined parity (even or odd) of a set of measurement results is consistent. 
    If noise or errors disrupt the circuit, the Detector can identify this through unexpected changes in parity, signaling potential measurement errors. 
    This functionality helps in error detection by revealing inconsistencies caused by disturbances.

    See Also:
        :class:`QubitCoordinates`
        :class:`ShiftCoordinates`
        :class:`ObservableInclude`
        :class:`Tick`

    Examples:

        Adding Detector operation to a Circuit (applied to classical bits):

        >>> from mimiqcircuits import *
        >>> op = Detector([0.1,0.9])
        >>> op
        lazy Detector(?, [0.1, 0.9])

        Fill the Lazy argument by calling the number of bits

        >>> op_filled = op(1)
        >>> op_filled
        Detector(0.1, 0.9)
        >>> op_filled.num_bits
        1

        Getting the nots

        >>> op_filled.get_notes()
        [0.1, 0.9]

        Define a new Operation

        >>> op = Detector(1, [0.5, 1.0])
        >>> op.get_notes()
        [0.5, 1.0]

        Add to the Circuit

        >>> c = Circuit()
        >>> c.push(Detector(1), 0)
        0-qubit circuit with 1 instructions:
        └── Detector() @ c[0]
        <BLANKLINE>
        >>> c.push(Detector(1, [0.5, 1.0]), 1)
        0-qubit circuit with 2 instructions:
        ├── Detector() @ c[0]
        └── Detector(0.5, 1.0) @ c[1]
        <BLANKLINE>
    """
    _name = "Detector"
    _num_zvars = 0
    _num_qubits = 0
    _num_cregs = 1
    _num_qregs = 0
    _num_zregs = 0

    def __new__(cls, *args):
        if len(args) == 0:
            return object.__new__(cls)
        elif len(args) == 1 and isinstance(args[0], int):
            return object.__new__(cls)
        elif len(args) == 2 and isinstance(args[0], int) and isinstance(args[1], list):
            return object.__new__(cls)
        elif len(args) == 1 and isinstance(args[0], list):
            return LazyExpr(cls, LazyArg(), args[0])
        elif len(args) > 1 and all(isinstance(arg, int) for arg in args):
            return object.__new__(cls)
        elif len(args) > 1 and all(isinstance(arg, float) for arg in args):
            return LazyExpr(cls, LazyArg(), args)
            
        else:
            raise ValueError("Incorrect Detector arguments")

    def __init__(self, *args):
        if len(args) == 0:
            N = 1
            notes = []
        elif len(args) == 1 and isinstance(args[0], int):
            N = args[0]
            notes = []
        elif len(args) == 2 and isinstance(args[0], int) and isinstance(args[1], list):
            N, notes = args
        elif len(args) > 1 and all(isinstance(arg, int) for arg in args):
            N, notes = args[0], list(args[1:])
        else:
            raise ValueError("Incorrect Detector arguments")

        if not isinstance(N, int) or N <= 0:
            raise ValueError("Detectors should be applied to at least 1 classical bit.")

        self._num_bits = N
        self.notes = [float(note) for note in (notes or [])]
        self._cregsizes = [N]

    @staticmethod
    def opname():
        return "Detector"

    def get_notes(self):
        return self.notes

    def iswrapper(self):
        return False

    def __str__(self):
        sep = "," if getattr(self, "_compact", False) else ", "
        notes_str = sep.join(map(str, self.get_notes()))
        return f"{self.opname()}({notes_str})"


class QubitCoordinates(AbstractAnnotation):
    r"""QubitCoordinates operation for specifying qubit positions.

    An annotation class used to specify the spatial location of a qubit in a quantum circuit. 
    Coordinates do not affect simulation results but are useful for visualizing and organizing qubit layouts within the circuit.

    See Also:
        :class:`Detector`
        :class:`ShiftCoordinates`
        :class:`ObservableInclude`
        :class:`Tick`

        
    Examples:

        Adding QubitCoordinates to a circuit:

        >>> from mimiqcircuits import *
        >>> QubitCoordinates([0.5, 0.75, 1.0])
        QubitCoordinates(0.5, 0.75, 1.0)
        >>> op = QubitCoordinates([0.5, 0.75, 1.0])
        >>> op.get_notes()
        [0.5, 0.75, 1.0]
        >>> c= Circuit()
        >>> c.push(QubitCoordinates([0.2, 0.3]), 0)
        1-qubit circuit with 1 instructions:
        └── QubitCoordinates(0.2, 0.3) @ q[0]
        <BLANKLINE>
    """
    _name = "QubitCoordinates"
    _num_zvars = 0
    _num_cregs = 0
    _num_qregs = 1
    _num_zregs = 0
    _num_bits = 0

    def __new__(cls, *args):
        if len(args) == 0:
            return LazyExpr(cls, LazyArg(), *args)
        if len(args) == 1 and isinstance(args[0], list):
            return object.__new__(cls)
        else:
            raise ValueError("Should provide a list of float")

    def __init__(self, coordinates: list):
        self._num_qubits = 1
        self.coordinates = [float(coord) for coord in (coordinates or [])]
        self._qregsizes = [1]

    @staticmethod
    def opname():
        return "QubitCoordinates"

    def get_notes(self):
        return self.coordinates

    def iswrapper(self):
        return False

    def __str__(self):
        return f"{self.opname()}({', '.join(map(str, self.get_notes()))})"


class ShiftCoordinates(AbstractAnnotation):
    r"""ShiftCoordinates operation for shifting the coordinates of qubits.

    An annotation class used to apply a shift to the spatial coordinates of subsequent qubit or detector annotations in a quantum circuit. 
    `ShiftCoordinates` accumulates offsets that adjust the position of related circuit components, aiding in visualization without affecting the simulation.

    See Also:
        :class:`Detector`
        :class:`QubitCoordinates`
        :class:`ObservableInclude`
        :class:`Tick`


    Examples:

        Adding ShiftCoordinates to represent spatial shifts in a circuit:

        >>> from mimiqcircuits import *
        >>> ShiftCoordinates(0.1, 0.2, 0.3)
        ShiftCoordinates(0.1, 0.2, 0.3)
        >>> ShiftCoordinates([0.4, 0.5])
        ShiftCoordinates(0.4, 0.5)
        >>> ShiftCoordinates(0.4, 0.5)
        ShiftCoordinates(0.4, 0.5)
        >>> op = ShiftCoordinates(0.1, 0.9)
        >>> op.get_notes()
        [0.1, 0.9]
        >>> c = Circuit()
        >>> c.push(ShiftCoordinates(0.4, 0.2))
        0-qubit circuit with 1 instructions:
        └── ShiftCoordinates(0.4, 0.2)
        <BLANKLINE>
    """
    _name = "ShiftCoordinates"
    _num_zvars = 0
    _num_cregs = 0
    _num_qregs = 0
    _num_zregs = 0

    def __new__(cls, *args):
        if len(args) == 0:
            return LazyExpr(cls, LazyArg(), *args)
        elif len(args) == 1 and isinstance(args[0], list):
            return object.__new__(cls)

        elif all(isinstance(arg, (int, float)) for arg in args):
            return object.__new__(cls)
        else:
            raise ValueError("Should provide a list of float")

    def __init__(self, *coordinates):
        self._num_qubits = 0
        self._num_bits = 0
        self._num_zvars = 0
        # If a list is provided as a single argument, use it directly
        if len(coordinates) == 1 and isinstance(coordinates[0], list):
            self.coordinates = [float(coord) for coord in coordinates[0]]
        # Otherwise, assume multiple float arguments are provided
        else:
            self.coordinates = [float(coord) for coord in coordinates]

    @staticmethod
    def opname():
        return "ShiftCoordinates"

    def get_notes(self):
        return self.coordinates

    def iswrapper(self):
        return False

    def __str__(self):
        return f"{self.opname()}({', '.join(map(str, self.get_notes()))})"


class ObservableInclude(AbstractAnnotation):
    r"""ObservableInclude operation for including observable index.

    An annotation class for adding measurement records to a specified logical observable within a quantum circuit. 
    Observables are sets of measurements expected to produce a deterministic result, used to track specific logical qubit states across operations.

    The ObservableInclude class tags a group of measurement records as a logical observable, 
    representing a consistent, predictable result under noiseless conditions. 
    This grouping allows for tracking the state of logical qubits across circuit operations, which is crucial for error correction. 
    Logical observables monitor encoded qubit states by combining multiple measurements, providing robustness against noise 
    and helping to identify any deviations that indicate potential errors.

    See Also:
        :class:`Detector`
        :class:`QubitCoordinates`
        :class:`ShiftCoordinates`
        :class:`Tick`

    Examples:

        Adding ObservableInclude to a circuit:

        >>> from mimiqcircuits import *
        >>> ObservableInclude(2)
        ObservableInclude()
        >>> c= Circuit()
        >>> c.push(ObservableInclude(1), 0)
        0-qubit circuit with 1 instructions:
        └── ObservableInclude() @ c[0]
        <BLANKLINE>
    """
    _name = "ObservableInclude"
    _num_zvars = 0
    _num_qubits = 0
    _num_cregs = 1
    _num_qregs = 0
    _num_zregs = 0

    def __new__(cls, *args):
        if len(args) == 0:
            return object.__new__(cls)
        elif len(args) == 1 and isinstance(args[0], int):
            return object.__new__(cls)
        elif len(args) == 2 and isinstance(args[0], int) and isinstance(args[1], list):
            return object.__new__(cls)
        elif len(args) == 1 and isinstance(args[0], list):
            return LazyExpr(cls, LazyArg(), args[0])
        elif len(args) > 1 and all(isinstance(arg, int) for arg in args):
            return object.__new__(cls)
        else:
            raise ValueError("Incorrect ObservableInclude arguments")

    def __init__(self, *args):
        if len(args) == 0:
            N = 1
            notes = []
        elif len(args) == 1 and isinstance(args[0], int):
            N = args[0]
            notes = []
        elif len(args) == 2 and isinstance(args[0], int) and isinstance(args[1], list):
            N, notes = args
        elif len(args) > 1 and all(isinstance(arg, int) for arg in args):
            N, notes = args[0], list(args[1:])
        else:
            raise ValueError("Incorrect ObservableInclude arguments")

        if not isinstance(N, int) or N <= 0:
            raise ValueError("ObservableIncludes should be applied to at least 1 classical bit.")

        self._num_bits = N
        self.notes = [int(note) for note in notes]
        self._cregsizes = [N]

    @staticmethod
    def opname():
        return "ObservableInclude"

    def get_notes(self):
        return self.notes

    def iswrapper(self):
        return False

    def __str__(self):
        sep = "," if getattr(self, "_compact", False) else ", "
        notes_str = sep.join(map(str, self.get_notes()))
        return f"{self.opname()}({notes_str})"


class Tick(AbstractAnnotation):
    """
    An annotation class representing a timing marker or layer boundary in a quantum circuit.
    `Tick` does not affect simulation but provides structure by separating operations into distinct time steps,
    which is useful for visualization and analysis.

    See Also:
        :class:`Detector`
        :class:`QubitCoordinates`
        :class:`ShiftCoordinates`
        :class:`ObservableInclude`

    Examples:
        >>> from mimiqcircuits import *
        >>> tick = Tick()
        >>> print(tick)
        Tick

        >>> c = Circuit()
        >>> c.push(tick)
        0-qubit circuit with 1 instructions:
        └── Tick
        <BLANKLINE>
    """
    _name = "Tick"
    _num_zvars = 0
    _num_cregs = 0
    _num_qregs = 0
    _num_zregs = 0
    _num_bits = 0
    _num_qubits = 0

    @staticmethod
    def opname():
        return "Tick"
    
    def get_notes(self):
        return []

    def iswrapper(self):
        return False

    def __str__(self):
        return f"{self.opname()}"

    def __repr__(self):
        return self.__str__()


__all__ = ["Detector", "QubitCoordinates", "ShiftCoordinates", "ObservableInclude"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.operation import Operation
import mimiqcircuits as mc
from mimiqcircuits.canvas import _find_unit_range, _string_with_square


class IfStatement(Operation):
    """Conditional operation

    This operation applies a specific operation if a given classical bit condition is met.

    Examples:
        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(IfStatement(GateX(), BitString('1')), 0, 0)
        1-qubit circuit with 1 instructions:
        └── IF (c==1) X @ q[0], c[0]
        <BLANKLINE>
    """

    _name = "If"

    _num_qubits = None

    _num_bits = None
    _num_cregs = 1

    _num_zvars = 0

    _op = None
    _bitstring = None

    def __init__(self, operation, bitstring: mc.BitString):
        if isinstance(operation, type) and issubclass(operation, Operation):
            op = operation()
        elif isinstance(operation, Operation):
            op = operation
        else:
            raise ValueError("Operation must be an Operation object or type.")

        if not isinstance(bitstring, mc.BitString):
            raise ValueError("bitstring must be a BitString object.")

        super().__init__()

        self._num_qubits = op.num_qubits
        self._num_qregs = op.num_qregs
        self._qregsizes = op.qregsizes

        self._num_bits = len(bitstring)
        self._cregsizes = [self._num_bits]

        self._op = op
        self._bitstring = bitstring

    @property
    def op(self):
        return self._op

    @op.setter
    def op(self, op):
        raise ValueError("Cannot set op. Read only parameter.")

    @property
    def bitstring(self):
        return self._bitstring

    def iswrapper(self):
        return True

    def getparams(self):
        return self.op.getparams()

    def inverse(self):
        raise NotImplementedError("Inverse not implemented for IfStatement")

    def power(self, power):
        raise NotImplementedError("Power not implemented for IfStatement")

    def control(self, num_controls):
        raise NotImplementedError("Control not implemented for IfStatement")

    def __str__(self):
        return f"IF (c=={self.bitstring.to01()}) {self.op}"

    def evaluate(self, d):
        return IfStatement(self.op.evaluate(d), self.bitstring)

    def get_operation(self):
        return self.op

    def get_bitstring(self):
        return self.bitstring

    def asciiwidth(self, qubits, bits, zvars):
        val = self.get_bitstring()
        gw = self.op.asciiwidth(qubits, [], [])
        bstr = _string_with_square(_find_unit_range(bits), ",")
        iw = len(f"c{bstr}==" + val.to01()) + 2

        return max(gw, iw)

    def _decompose(self, circuit, qtargets, ctargets, ztargets):
        decomposed_insts = self.op.decompose()
        bs = self.get_bitstring()
        for inst in decomposed_insts:
            conditional_operation = IfStatement(inst.operation, bs)
            targeted_qubits = [qtargets[i] for i in inst.get_qubits()]
            circuit.push(conditional_operation, *targeted_qubits, *ctargets, *ztargets)

        return circuit


__all__ = ["IfStatement"]



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import mimiqcircuits as mc
import numpy as np
from itertools import product
import symengine as se
from mimiqcircuits.operations.krauschannel import krauschannel
import sympy as sp


class Depolarizing(krauschannel):
    r"""N-qubit depolarizing noise channel.

    The Kraus operators for the depolarizing channel are given by:

    .. math::
        E_1 = \sqrt{1-p} I_N, \quad E_i = \sqrt{p/(4^N-1)} P_i,

    where :math:`p \in [0,1]` is a probability, and :math:`P_i` is an :math:`N`-qubit Pauli string
    operator, i.e., a tensor product of one-qubit Pauli operators.

    There is exactly one Kraus operator :math:`E_{i>1}` for each distinct combination of
    Pauli operators :math:`P_i`, except for the :math:`N`-qubit identity
    :math:`I_N = I \otimes I \otimes I \otimes \dots`.

    For example:

    - For one qubit, we have 3 operators :math:`P_i \in \{X, Y, Z\}`.
    - For two qubits, we have 15 operators :math:`P_i \in \{ I\otimes X, I\otimes Y,
      I\otimes Z, X\otimes I, Y\otimes I, Z\otimes I, X\otimes X, X\otimes Y, X\otimes Z,
      Y\otimes X, Y\otimes Y, Y\otimes Z, Z\otimes X, Z\otimes Y, Z\otimes Z \}`.

    This channel is a mixed unitary channel (see :func:`ismixedunitary`),
    and is a special case of :class:`PauliNoise`.

    See Also:
        :class:`PauliString`
        :class:`PauliNoise`

    Parameters:
        N (int): Number of qubits.
        p (float): Probability of error, i.e., the probability of not applying the identity.

    Examples:
        Depolarizing channels can be defined for any number of qubits:

        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(Depolarizing(1, 0.1), 1)
        2-qubit circuit with 1 instructions:
        └── Depolarizing(0.1) @ q[1]
        <BLANKLINE>

        >>> c.push(Depolarizing(5, 0.1), 1, 2, 3, 4, 5)
        6-qubit circuit with 2 instructions:
        ├── Depolarizing(0.1) @ q[1]
        └── Depolarizing(0.1) @ q[1,2,3,4,5]
        <BLANKLINE>

        For one and two qubits, you can use the shorthand notation:

        >>> c.push(Depolarizing(1, 0.1), 1)
        6-qubit circuit with 3 instructions:
        ├── Depolarizing(0.1) @ q[1]
        ├── Depolarizing(0.1) @ q[1,2,3,4,5]
        └── Depolarizing(0.1) @ q[1]
        <BLANKLINE>

        >>> c.push(Depolarizing(2, 0.1), 1, 2)
        6-qubit circuit with 4 instructions:
        ├── Depolarizing(0.1) @ q[1]
        ├── Depolarizing(0.1) @ q[1,2,3,4,5]
        ├── Depolarizing(0.1) @ q[1]
        └── Depolarizing(0.1) @ q[1,2]
        <BLANKLINE>
    """

    _name = "Depolarizing"
    _parnames = ()
    _num_qubits = None
    _parnames = ()

    def __init__(self, N, p):
        if N < 1:
            raise ValueError("Cannot define a 0-qubit depolarizing noise channel")
        if not isinstance (p, (se.Basic, sp.Basic)) and (p < 0 or p > 1):
            raise ValueError("Probability p needs to be between 0 and 1.")
        self.N = N
        self.p = p
        self._num_qubits = self.N
        self._parnames = ("N", "p")

    @property
    def parnames(self):
        return self._parnames

    @property
    def num_qubits(self):
        return self._num_qubits

    def evaluate(self, d={}):
        if hasattr(self.p, 'subs'):
            substituted_p = self.p.subs(d)
            evaluated_p = float(substituted_p.evalf()) if substituted_p.is_number else substituted_p
        else:
            evaluated_p = self.p

        if isinstance(evaluated_p, (float, int)) and not (0 <= evaluated_p <= 1):
            raise ValueError("Probability after evaluation must be between 0 and 1")

        return Depolarizing(self.N, evaluated_p)

    def probabilities(self):
        N = self.N
        return np.concatenate(([1 - self.p], np.repeat(self.p / (4**N - 1), 4**N - 1)))

    def unitarygates(self):
        N = self.N
        paulis = ["I", "X", "Y", "Z"]
        combinations = ["".join(p) for p in product(paulis, repeat=N)]
        return [mc.PauliString(comb) for comb in combinations]

    def unitarymatrices(self):
        return [gate.matrix() for gate in self.unitarygates()]

    def krausmatrices(self):
        probabilities = self.probabilities()
        unitary_matrices = self.unitarymatrices()
        sqrt_probs = [se.sqrt(p) for p in probabilities]
        return [
            sqrt_p * u_matrix for sqrt_p, u_matrix in zip(sqrt_probs, unitary_matrices)
        ]

    def krausoperators(self):
        return [mc.Operator(Ek) for Ek in self.krausmatrices()]

    def iswrapper(self):
        pass

    @classmethod
    def ismixedunitary(self):
        return True

    def __str__(self):
        return f"{self._name}({self.p})"


class Depolarizing1(krauschannel):
    def __new__(self, p):
        return Depolarizing(1, p)


class Depolarizing2(krauschannel):
    def __new__(self, p):
        return Depolarizing(2, p)


__all__ = ["Depolarizing", "Depolarizing1", "Depolarizing2"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc


class GateCCP(mc.Control):
    r"""Three qubit Controlled-Controlled-Phase gate.

    By convention, the first two qubits are the controls and the third is the
    target

    Arguments:
        lmbda: Phase angle.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> lmbda = Symbol('lmbda')
        >>> GateCCP(lmbda), GateCCP(lmbda).num_controls, GateCCP(lmbda).num_targets, GateCCP(lmbda).num_qubits
        (C₂P(lmbda), 2, 1, 3)
        >>> GateCCP(lmbda).matrix()
        [1.0, 0, 0, 0, 0, 0, 0, 0]
        [0, 1.0, 0, 0, 0, 0, 0, 0]
        [0, 0, 1.0, 0, 0, 0, 0, 0]
        [0, 0, 0, 1.0, 0, 0, 0, 0]
        [0, 0, 0, 0, 1.0, 0, 0, 0]
        [0, 0, 0, 0, 0, 1.0, 0, 0]
        [0, 0, 0, 0, 0, 0, 1.0, 0]
        [0, 0, 0, 0, 0, 0, 0, exp(I*lmbda)]
        <BLANKLINE>
        >>> c = Circuit().push(GateCCP(lmbda), 0, 1, 2)
        >>> c
        3-qubit circuit with 1 instructions:
        └── C₂P(lmbda) @ q[0,1], q[2]
        <BLANKLINE>
        >>> GateCCP(lmbda).power(2), GateCCP(lmbda).inverse()
        (C₂P(2*lmbda), C₂P(-lmbda))
        >>> GateCCP(lmbda).decompose()
        3-qubit circuit with 5 instructions:
        ├── CP((1/2)*lmbda) @ q[1], q[2]
        ├── CX @ q[0], q[1]
        ├── CP((-1/2)*lmbda) @ q[1], q[2]
        ├── CX @ q[0], q[1]
        └── CP((1/2)*lmbda) @ q[0], q[2]
        <BLANKLINE>
    """

    def __init__(self, lmbda):
        super().__init__(2, mc.GateP(lmbda))

    def _decompose(self, circ, qubits, bits, zvars):
        c1, c2, t = qubits
        lmbda = self.op.lmbda
        circ.push(mc.GateCP(lmbda / 2), c2, t)
        circ.push(mc.GateCX(), c1, c2)
        circ.push(mc.GateCP(-lmbda / 2), c2, t)
        circ.push(mc.GateCX(), c1, c2)
        circ.push(mc.GateCP(lmbda / 2), c1, t)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from functools import reduce
from mimiqcircuits.operations.gates.gate import Gate
import symengine as se
import sympy as sp
import mimiqcircuits.lazy as lz
import mimiqcircuits as mc


def to_superscript(number):
    superscript_map = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
    return str(number).translate(superscript_map)


class Parallel(Gate):
    """Parallel operation

    This is a composite operation that applies multiple gates in parallel to the circuit at once.

    Examples:
        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Parallel(3,GateX()),1,2,3)
        4-qubit circuit with 1 instructions:
        └── ⨷ ³ X @ q[1], q[2], q[3]
        <BLANKLINE>
    """

    _name = "Parallel"
    _num_qubits = None
    _num_bits = 0
    _num_repeats = None
    _op = None

    def __init__(self, num_repeats, op: Gate):
        if not isinstance(op, (Gate)):
            raise TypeError(f"{op.__class__.__name__} cannot be Paralleled.")

        if self.num_bits != 0:
            raise ValueError("Parallel operations cannot act on classical bits.")

        if num_repeats < 2:
            raise ValueError(
                f"Parallel operations must have at least two repeats. If one, just use {op}."
            )

        super().__init__()
        self._num_qubits = op.num_qubits * num_repeats
        self._num_repeats = num_repeats
        self._op = op
        self._qregsizes = [1] * self._num_qubits
        self._num_qregs = self._num_qubits

    def _matrix(self):
        op_matrix = self.op.matrix()
        return reduce(np.kron, [op_matrix] * (self._num_repeats))

    @property
    def num_repeats(self):
        return self._num_repeats

    @num_repeats.setter
    def num_repeats(self, value):
        raise ValueError("Cannot set num_repeats. Read only parameter.")

    @property
    def op(self):
        return self._op

    @op.setter
    def op(self, op):
        raise ValueError("Cannot set op. Read only parameter.")

    def iswrapper(self):
        return True

    def _getparams(self):
        return self.op.getparams()

    def inverse(self):
        return Parallel(self.num_repeats, self.op.inverse())

    def power(self, *args):
        if len(args) == 0:
            return lz.power(self)
        elif len(args) == 1:
            pwr = args[0]
            return mc.Power(self.op, pwr).parallel(self.num_repeats)
        else:
            raise ValueError("Invalid number of arguments.")

    def control(self, *args):
        if len(args) == 0:
            return lz.control(self)
        elif len(args) == 1:
            num_controls = args[0]
            return mc.Control(num_controls, self)
        else:
            raise ValueError("Invalid number of arguments.")

    def parallel(self, *args):
        if len(args) == 0:
            return lz.parallel(self)
        elif len(args) == 1:
            num_repeats = args[0]
            return Parallel(self.num_repeats * num_repeats, self.op)
        else:
            raise ValueError("Invalid number of arguments.")

    def _decompose(self, circ, qubits, bits, zvars):
        nq = self.op.num_qubits
        for i in range(self.num_repeats):
            q = [qubits[j] for j in range(i * nq, (i + 1) * nq)]
            circ.push(self.op, *q)
        return circ

    def __str__(self):
        tail = f"⨷ {to_superscript(self.num_repeats)}"
        return f"{tail} {self.op}"

    def evaluate(self, d):
        repeat = self.num_repeats
        return self.op.evaluate(d).parallel(repeat)

    def get_operation(self):
        return self.op


# export operations
__all__ = ["Parallel"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fractions import Fraction
from symengine import pi
from mimiqcircuits.printutils import print_wrapped_parens
import sympy as sp
import numpy as np
import mimiqcircuits as mc
from mimiqcircuits.operations.gates.gate import Gate


class Power(Gate):
    """Power operation.

    Represents a Power operation raised to a specified exponent.

    Examples:
        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Power(GateX(),1/2),1)
        2-qubit circuit with 1 instructions:
        └── SX @ q[1]
        <BLANKLINE>
        >>> c.push(Power(GateX(),5),1)
        2-qubit circuit with 2 instructions:
        ├── SX @ q[1]
        └── X**5 @ q[1]
        <BLANKLINE>
        >>> c.decompose()
        2-qubit circuit with 6 instructions:
        ├── U(1.5707963267948966, -1.5707963267948966, 1.5707963267948961, 0.7853981633974482) @ q[1]
        ├── X @ q[1]
        ├── X @ q[1]
        ├── X @ q[1]
        ├── X @ q[1]
        └── X @ q[1]
        <BLANKLINE>
    """

    _name = "Power"

    _num_qubits = None

    _num_bits = 0
    _num_cregs = 0

    _op = None
    _parnames = ("exponent",)

    def __init__(self, operation, exponent, *args, **kwargs):
        if isinstance(operation, type) and issubclass(operation, mc.Gate):
            op = operation(*args, **kwargs)
        elif isinstance(operation, mc.Gate):
            op = operation
        else:
            raise ValueError("Operation must be an Gate object or type.")

        if self.num_bits != 0:
            raise ValueError("Power operation cannot act on classical bits.")

        super().__init__()

        self._exponent = exponent
        self._op = op
        self._num_qubits = op.num_qubits
        self._num_qregs = op.num_qregs
        self._qregsizes = op.qregsizes
        self._parnames = op.parnames
        if isinstance(op, mc.Power):
            self._op = op.op
            self._exponent = exponent * op.exponent

    @property
    def op(self):
        return self._op

    @op.setter
    def op(self, op):
        raise ValueError("Cannot set op. Read only parameter.")

    @property
    def exponent(self):
        return self._exponent

    @exponent.setter
    def exponent(self, power):
        raise ValueError("Cannot set exponent. Read only parameter.")

    def iswrapper(self):
        return True

    def _power(self, pwr):
        return self.op.power(pwr * self._exponent)

    def power(self, *args):
        if len(args) == 0:
            return mc.power(self)
        elif len(args) == 1:
            exponent = args[0]
            return self._power(exponent)
        else:
            raise ValueError("Invalid number of arguments.")

    def __pow__(self, exponent):
        return self.power(exponent)

    def inverse(self):
        return mc.Inverse(self)

    def control(self, *args):
        if len(args) == 0:
            return mc.control(self)
        elif len(args) == 1:
            num_controls = args[0]
            return mc.Control(num_controls, self)
        else:
            raise ValueError("Invalid number of arguments.")

    def parallel(self, *args):
        if len(args) == 0:
            return mc.parallel(self)
        elif len(args) == 1:
            num_repeats = args[0]
            return mc.Parallel(num_repeats, self)
        else:
            raise ValueError("Invalid number of arguments.")

    def _matrix(self):
        matrix = sp.Matrix(self.op.matrix().tolist())
        pow_matrix = matrix ** (self.exponent)
        return pow_matrix

    def getparams(self):
        return self.op.getparams()

    def __str__(self):
        if self._op == mc.GateX() and self._exponent == 1 / 2:
            return mc.GateSX().name

        fraction = Fraction(self.exponent).limit_denominator(100)

        if float(fraction) == self.exponent and int(fraction) != self.exponent:
            return f"{print_wrapped_parens(self.op)}**({fraction})"

        if self.exponent == pi:
            return f"{print_wrapped_parens(self.op)}**π"
        if self.exponent == -pi:
            return f"{print_wrapped_parens(self.op)}**(-π)"

        divpi = Fraction(self.exponent / np.pi).limit_denominator(100)

        if float(divpi) == self.exponent / np.pi:
            if divpi == 1:
                return f"{print_wrapped_parens(self.op)}**π"

            if divpi == -1:
                return f"{print_wrapped_parens(self.op)}**(-1)"

            if divpi > 0:
                return f"{print_wrapped_parens(self.op)}**({divpi} * π)"

            return f"{print_wrapped_parens(self.op)}**(({divpi}) * π)"

        if self.exponent < 0:
            return f"{print_wrapped_parens(self.op)}**({self.exponent})"

        return f"{print_wrapped_parens(self.op)}**{self.exponent}"

    def __repr__(self):
        return self.__str__()

    def evaluate(self, d):
        exponent = self.exponent
        return self.op.evaluate(d).power(exponent)

    def _decompose(self, circ, qubits, bits, zvars):
        if isinstance(self.exponent, int) and self.exponent >= 1:
            for _ in range(self.exponent):
                circ.push(self.op, *qubits)
            return circ

        # try to decompose,
        # if there is only a gate, maybe it is ok
        # if the gates are all diagonal then we can continue
        # otherwise just do nothing and push the same thing
        cop = self.op._decompose(mc.Circuit(), qubits, bits, zvars)

        if len(cop) == 1:
            circ.push(cop.instructions[0].operation._power(self.exponent), *qubits)
            return circ

        circ.push(self, *qubits)

        return circ

    def decompose(self):
        return self._decompose(
            mc.Circuit(),
            range(self.num_qubits),
            range(self.num_bits),
            range(self.num_zvars),
        )


# export operation
__all__ = ["Power"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.krauschannel import krauschannel
import mimiqcircuits as mc


def _check_normaliz_length(normaliz, expected_length):
    if len(normaliz) != expected_length:
        raise ValueError(f"normaliz must be a list of length {expected_length}.")


class Reset(krauschannel):
    """Reset operation.

    Quantum operation that resets the status of one qubit to the :math:`\\ket{0}` state.

    .. warning::
        This operation is non-reversible.

    Examples:
        Adding Reset operation to the Circuit (The args can be: range, list, tuple, set or int)

        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Reset(), 0)
        1-qubit circuit with 1 instructions:
        └── Reset @ q[0]
        <BLANKLINE>
        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Reset(),(0,1,2))
        3-qubit circuit with 3 instructions:
        ├── Reset @ q[0]
        ├── Reset @ q[1]
        └── Reset @ q[2]
        <BLANKLINE>

    .. seealso::
        :class:`ResetX`, :class:`ResetY`, :class:`ResetZ` - Variants of the Reset operation along different axes.
    """

    _name = "Reset"
    _num_qubits = 1
    _num_bits = 0
    _qregsizes = [1]

    def inverse(self):
        raise TypeError("Reset is not inversible")

    def power(self, pwr):
        raise TypeError("Reset^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled Reset is not defined.")

    def iswrapper(self):
        return False

    def _matrix(self):
        raise AttributeError("Reset has no attribute 'matrix'.")

    def __str__(self):
        return f"{self._name}"

    def asciiwidth(self, qubits, bits, zvars):
        return 5

    def krausoperators(self, normaliz=None):
        if normaliz is None:
            normaliz = [1, 1]
        _check_normaliz_length(normaliz, 2)
        return [mc.Projector0(1 / normaliz[0]), mc.SigmaMinus(1 / normaliz[1])]

    def __str__(self):
        return self._name


class ResetX(krauschannel):
    """ResetX operation.

    This operation is performed by applying a Hadamard gate (H) to the qubit, which transforms the
    qubit from the computational basis (:math:`\\ket{0}` or :math:`\\ket{1}`) to the superposition
    basis (:math:`\\ket{+}` or :math:`\\ket{-}`). After applying the Hadamard gate, a standard Reset
    operation is performed, resetting the qubit to the :math:`\\ket{0}` state. Finally, another Hadamard
    gate is applied to convert the :math:`\\ket{0}` state back into the :math:`\\ket{+}` state.

    .. warning::
        This operation is non-reversible.

    Examples:
        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(ResetX(), 0)
        1-qubit circuit with 1 instructions:
        └── ResetX @ q[0]
        <BLANKLINE>
        >>> c.decompose()
        1-qubit circuit with 3 instructions:
        ├── H @ q[0]
        ├── Reset @ q[0]
        └── H @ q[0]
        <BLANKLINE>

    .. seealso::
        :class:`Reset`, :class:`ResetY`, :class:`ResetZ` - Variants of the Reset operation along different axes.
    """

    _name = "ResetX"
    _num_qubits = 1
    _num_bits = 0
    _num_zvars = 0
    _qregsizes = [1]

    def inverse(self):
        raise TypeError("ResetX is not inversible")

    def power(self, pwr):
        raise TypeError("ResetX^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled ResetX is not defined.")

    def iswrapper(self):
        return False

    def _matrix(self):
        raise AttributeError("ResetX has no attribute 'matrix'.")

    def __str__(self):
        return self._name

    def _decompose(self, circ: mc.Circuit, qtargets, bits=None, zvars=None):
        q = qtargets[0]
        circ.push(mc.GateH(), q)
        circ.push(Reset(), q)
        circ.push(mc.GateH(), q)
        return circ


class ResetY(krauschannel):
    """ResetY operation.

    The `ResetY` operation works by first rotating the qubit's state so that the Y-axis aligns with the Z-axis,
    applying a standard reset (which resets the qubit to the :math:`\\ket{0}` state), and then rotating the qubit
    back to its original basis.

    .. warning::
        This operation is non-reversible.

    Examples:
        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(ResetY(), 0)
        1-qubit circuit with 1 instructions:
        └── ResetY @ q[0]
        <BLANKLINE>
        >>> c.decompose()
        1-qubit circuit with 3 instructions:
        ├── HYZ @ q[0]
        ├── Reset @ q[0]
        └── HYZ @ q[0]
        <BLANKLINE>

    .. seealso::
        :class:`Reset`, :class:`ResetX`, :class:`ResetZ` - Variants of the Reset operation along different axes.
    """

    _name = "ResetY"
    _num_qubits = 1
    _num_bits = 0
    _qregsizes = [1]

    def inverse(self):
        raise TypeError("ResetY is not inversible")

    def power(self, pwr):
        raise TypeError("ResetY^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled ResetY is not defined.")

    def iswrapper(self):
        return False

    def _matrix(self):
        raise AttributeError("ResetY has no attribute 'matrix'.")

    def __str__(self):
        return self._name

    def _decompose(self, circ: mc.Circuit, qtargets, bits=None, zvars=None):
        q = qtargets[0]
        circ.push(mc.GateHYZ(), q)
        circ.push(Reset(), q)
        circ.push(mc.GateHYZ(), q)
        return circ


class ResetZ(krauschannel):
    """ResetZ operation.

    Quantum operation that resets the status of one qubit to the :math:`\\ket{0}` state along the Z-axis.
    This operation is an alias of the :class:`Reset` operation.

    .. warning::
        This operation is non-reversible.

    Examples:
        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(ResetZ(), 0)
        1-qubit circuit with 1 instructions:
        └── Reset @ q[0]
        <BLANKLINE>
        >>> c.decompose()
        1-qubit circuit with 1 instructions:
        └── Reset @ q[0]
        <BLANKLINE>

    .. seealso::
        :class:`Reset`, :class:`ResetX`, :class:`ResetY` - Variants of the Reset operation along different axes.
    """

    def __new__(self):
        return Reset()


# export operations
__all__ = ["Reset", "ResetX", "ResetY", "ResetZ"]



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from mimiqcircuits import Operation, BitString


class Amplitude(Operation):
    """Amplitude operation

    multi qubit Amplitude operation in the computational basis

    The operation projects the quantum states complex variables and stores in a z-register.

    Examples:
        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(Amplitude(BitString(2)),0)
        0-qubit circuit with 1 instructions:
        └── Amplitude(bs"00") @ z[0]
        <BLANKLINE>
    """

    _name = "Amplitude"
    _num_zvars = 1
    _num_bits = 0
    _num_zregs = 1
    _num_cregs = 0
    _num_qregs = 0
    _num_qubits = 0

    def __init__(self, bs: BitString):
        self.bs = bs

        super().__init__()
        self._zregsizes = [1]

    @property
    def zregsizes(self):
        return self._zregsizes

    def iswrapper(self):
        return False

    def _decompose(self, circ, qubits, bits, zvars):
        return circ.push(self, *qubits, *bits, *zvars)

    def inverse(self):
        raise NotImplementedError("Cannot inverse an Amplitude operation.")

    def _power(self, _):
        raise NotImplementedError("Cannot elevate an Amplitude operation to any power.")

    def __str__(self):
        return f'{self._name}(bs"{self.bs.to01()}")'
    
    @staticmethod
    def isunitary():
        return True

__all__ = ["Amplitude"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.u import GateU
from mimiqcircuits.operations.gates.standard.deprecated import GateU3
from mimiqcircuits.operations.utils import control_one_defined
import mimiqcircuits as mc
from symengine import pi, Matrix, cos, sin, I, exp


class GateRX(mcg.Gate):
    r"""Single qubit Rotation gate around the axis :math:`\hat{x}`

    **Matrix representation:**

    .. math::
        \operatorname{RX}(\theta) = \begin{pmatrix}
            \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
            -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
        \end{pmatrix}

    Parameters:
        theta: Rotation angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta = Symbol('theta')
        >>> GateRX(theta)
        RX(theta)
        >>> GateRX(theta).matrix()
        [cos((1/2)*theta), -I*sin((1/2)*theta)]
        [-I*sin((1/2)*theta), cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateRX(theta), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── RX(theta) @ q[0]
        <BLANKLINE>
        >>> GateRX(theta).power(2), GateRX(theta).inverse()
        (RX(2*theta), RX(-theta))
        >>> GateRX(theta).decompose()
        1-qubit circuit with 1 instructions:
        └── U(theta, (-1/2)*pi, (1/2)*pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "RX"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("theta",)

    def __init__(self, theta):
        self.theta = theta

    def _matrix(self):
        return Matrix(
            [
                [cos(self.theta / 2), -I * sin(self.theta / 2)],
                [-I * sin(self.theta / 2), cos(self.theta / 2)],
            ]
        )

    def inverse(self):
        return GateRX(-self.theta)

    def _power(self, p):
        return GateRX(self.theta * p)

    def isidentity(self):
        if self.theta == 0:
            return True
        return False

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCRX(self.theta))

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(self.theta, -pi / 2, pi / 2), q)
        return circ


class GateRY(mcg.Gate):
    r"""Single qubit Rotation gate around the axis :math:`\hat{y}`

    **Matrix representation:**

    .. math::
        \operatorname{RY}(\theta) = \begin{pmatrix}
            \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
            \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
        \end{pmatrix}

    Parameters:
        theta (float): Rotation angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta = Symbol('theta')
        >>> GateRY(theta)
        RY(theta)
        >>> GateRY(theta).matrix()
        [cos((1/2)*theta), -sin((1/2)*theta)]
        [sin((1/2)*theta), cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateRY(theta), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── RY(theta) @ q[0]
        <BLANKLINE>
        >>> GateRY(theta).power(2), GateRY(theta).inverse()
        (RY(2*theta), RY(-theta))
        >>> GateRY(theta).decompose()
        1-qubit circuit with 1 instructions:
        └── U(theta, 0, 0, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "RY"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("theta",)

    def __init__(self, theta):
        self.theta = theta

    def _matrix(self):
        return Matrix(
            [
                [cos(self.theta / 2), -sin(self.theta / 2)],
                [sin(self.theta / 2), cos(self.theta / 2)],
            ]
        )

    def inverse(self):
        return GateRY(-self.theta)

    def _power(self, p):
        return GateRY(self.theta * p)

    def isidentity(self):
        if self.theta == 0:
            return True
        return False

    def _control(self, n):
        control_one_defined(n, self, mc.GateCRY(self.theta))

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(self.theta, 0, 0), q)
        return circ


class GateRZ(mcg.Gate):
    r"""Single qubit Rotation gate around the axis :math:`\hat{z}`

    **Matrix representation:**

    .. math::
        \operatorname{RZ}(\lambda) = \begin{pmatrix}
            e^{-i\frac{\lambda}{2}} & 0 \\
            0 & e^{i\frac{\lambda}{2}}
        \end{pmatrix}

    Parameters:
        lambda: Rotation angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> lmbda = Symbol('lambda')
        >>> GateRZ(lmbda)
        RZ(lambda)
        >>> GateRZ(lmbda).matrix()
        [exp(-1/2*I*lambda), 0]
        [0, exp(1/2*I*lambda)]
        <BLANKLINE>
        >>> c = Circuit().push(GateRZ(lmbda), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── RZ(lambda) @ q[0]
        <BLANKLINE>
        >>> GateRZ(lmbda).power(2), GateRZ(lmbda).inverse()
        (RZ(2*lambda), RZ(-lambda))
        >>> GateRZ(lmbda).decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, lambda, (-1/2)*lambda) @ q[0]
        <BLANKLINE>
    """

    _name = "RZ"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("lmbda",)

    def __init__(self, lmbda):
        self.lmbda = lmbda

    def _matrix(self):
        return Matrix([[exp(-I * self.lmbda / 2), 0], [0, exp(I * self.lmbda / 2)]])

    def inverse(self):
        return GateRZ(-self.lmbda)

    def _power(self, p):
        return GateRZ(self.lmbda * p)

    def isidentity(self):
        if self.lmbda == 0:
            return True
        return False

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCRZ(self.lmbda))

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(0, 0, self.lmbda, -self.lmbda / 2), q)
        return circ


class GateR(mcg.Gate):
    r"""Single qubit Rotation gate around the axis :math:`\cos(\phi)\hat{x} + \sin(\phi)\hat{y}`.

    **Matrix representation:**

    .. math::
        \operatorname R(\theta,\phi) = \begin{pmatrix}
          \cos \frac{\theta}{2}  & -i e^{-i\phi} \sin \frac{\theta}{2} \\
          -i e^{i \phi} \sin \frac{\theta}{2}  &  \cos \frac{\theta}{2}
        \end{pmatrix}

    Parameters:
        theta (float): The rotation angle in radians.
        phi (float): The axis of rotation in radians.

    Example:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta, phi = symbols('theta phi')
        >>> GateR(theta, phi)
        R(theta, phi)
        >>> GateR(theta, phi).matrix()
        [cos((1/2)*theta), -I*exp(-I*phi)*sin((1/2)*theta)]
        [-I*exp(I*phi)*sin((1/2)*theta), cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateR(theta, phi), 0)
        >>> GateR(theta, phi).power(2), GateR(theta, phi).inverse()
        (R(2*theta, phi), R(-theta, phi))
        >>> GateR(theta, phi).decompose()
        1-qubit circuit with 1 instructions:
        └── U3(theta, phi + (-1/2)*pi, -phi + (1/2)*pi) @ q[0]
        <BLANKLINE>
    """

    _name = "R"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("theta", "phi")

    def __init__(self, theta, phi):
        self.theta = theta
        self.phi = phi

    def _matrix(self):
        return Matrix(
            [
                [cos(self.theta / 2), -I * exp(-I * self.phi) * sin(self.theta / 2)],
                [-I * exp(I * self.phi) * sin(self.theta / 2), cos(self.theta / 2)],
            ]
        )

    def inverse(self):
        return GateR(-self.theta, self.phi)

    def _power(self, p):
        return GateR(self.theta * p, self.phi)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU3(self.theta, self.phi - pi / 2, -self.phi + pi / 2), q)
        return circ



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import symengine as se
import mimiqcircuits as mc


def to_subscript(number):
    subscript_map = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
    return str(number).translate(subscript_map)


class Projector0(mc.AbstractOperator):
    r"""One-qubit operator corresponding to a projection onto :math:`|0\rangle`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            a & 0 \\
            0 & 0
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.
    Equivalent to :class:`DiagonalOp(a, 0)`.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`Projector1`, :class:`DiagonalOp`

    Parameters:
        a (complex, optional): The top-left entry of the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> Projector0()
        P₀(1)

        >>> Projector0(0.5)
        P₀(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Projector0()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨P₀(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "Projector0"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return se.Matrix([[self.a, 0], [0, 0]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"P{to_subscript(0)}({self.a})"

    def rescale(self, scale):
        return Projector0(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return Projector0(abs(self.a) ** 2)


class Projector1(mc.AbstractOperator):
    r"""One-qubit operator corresponding to a projection onto :math:`|1\rangle`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            0 & 0 \\
            0 & a
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.
    Equivalent to :class:`DiagonalOp(0, a)`.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`Projector0`, :class:`DiagonalOp`

    Parameters:
        a (complex, optional): The bottom-right entry of the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> Projector1()
        P₁(1)

        >>> Projector1(0.5)
        P₁(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Projector1()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨P₁(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "Projector1"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return se.Matrix([[0, 0], [0, self.a]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"P{to_subscript(1)}({self.a})"

    def rescale(self, scale):
        return Projector1(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return Projector1(abs(self.a) ** 2)


class ProjectorX0(mc.AbstractOperator):
    r"""One-qubit operator corresponding to a projection onto :math:`|+\rangle`.

    **Matrix Representation**

    .. math::
        \frac{a}{2}
        \begin{pmatrix}
            1 & 1 \\
            1 & 1
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`ProjectorX1`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> ProjectorX0()
        PX₀(1)

        >>> ProjectorX0(0.5)
        PX₀(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(ProjectorX0()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨PX₀(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "PX0"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return self.a / 2 * se.Matrix([[1, 1], [1, 1]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"PX{to_subscript(0)}({self.a})"

    def rescale(self, scale):
        return ProjectorX0(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return ProjectorX0(abs(self.a) ** 2)


class ProjectorX1(mc.AbstractOperator):
    r"""One-qubit operator corresponding to a projection onto :math:`|-\rangle`.

    **Matrix Representation**

    .. math::
        \frac{a}{2}
        \begin{pmatrix}
            1 & -1 \\
            -1 & 1
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`ProjectorX0`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> ProjectorX1()
        PX₁(1)

        >>> ProjectorX1(0.5)
        PX₁(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(ProjectorX1()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨PX₁(1)⟩ @ q[1], z[1]
        <BLANKLINE>
        
    """

    _name = "PX1"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return self.a / 2 * se.Matrix([[1, -1], [-1, 1]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"PX{to_subscript(1)}({self.a})"

    def rescale(self, scale):
        return ProjectorX1(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return ProjectorX1(abs(self.a) ** 2)


class ProjectorY0(mc.AbstractOperator):
    r"""One-qubit operator corresponding to a projection onto :math:`|y+\rangle`.

    **Matrix Representation**

    .. math::
        \frac{a}{2}
        \begin{pmatrix}
            1 & -i \\
            i & 1
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`ProjectorY1`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> ProjectorY0()
        PY₀(1)

        >>> ProjectorY0(0.5)
        PY₀(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(ProjectorY0()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨PY₀(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "PY0"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return self.a / 2 * se.Matrix([[1, -1j], [1j, 1]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"PY{to_subscript(0)}({self.a})"

    def rescale(self, scale):
        return ProjectorY0(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return ProjectorY0(abs(self.a) ** 2)


class ProjectorY1(mc.AbstractOperator):
    r"""One-qubit operator corresponding to a projection onto :math:`|y-\rangle`.

    **Matrix Representation**

    .. math::
        \frac{a}{2}
        \begin{pmatrix}
            1 & i \\
            -i & 1
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`ProjectorY0`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> ProjectorY1()
        PY₁(1)

        >>> ProjectorY1(0.5)
        PY₁(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(ProjectorY1()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨PY₁(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "PY1"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return self.a / 2 * se.Matrix([[1, 1j], [-1j, 1]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"PY{to_subscript(1)}({self.a})"

    def rescale(self, scale):
        return ProjectorY1(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return ProjectorY1(abs(self.a) ** 2)


class Projector00(mc.AbstractOperator):
    r"""Two-qubit operator corresponding to a projection onto :math:`|00\rangle`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            a & 0 & 0 & 0\\
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & 0
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`Projector01`, :class:`Projector10`, :class:`Projector11`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> Projector00()
        P₀₀(1)

        >>> Projector00(0.5)
        P₀₀(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Projector00()), 1, 2, 1)
        3-qubit circuit with 1 instructions:
        └── ⟨P₀₀(1)⟩ @ q[1,2], z[1]
        <BLANKLINE>
    """

    _name = "Projector00"
    _num_qubits = 2
    _parnames = ()
    _qregsizes = [2]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return se.Matrix([[self.a, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"P{to_subscript(0)}{to_subscript(0)}({self.a})"

    def rescale(self, scale):
        return Projector00(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return Projector00(abs(self.a) ** 2)


class Projector01(mc.AbstractOperator):
    r"""Two-qubit operator corresponding to a projection onto :math:`|01\rangle`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            0 & 0 & 0 & 0\\
            0 & a & 0 & 0\\
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & 0
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`Projector00`, :class:`Projector10`, :class:`Projector11`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> Projector01()
        P₀₁(1)

        >>> Projector01(0.5)
        P₀₁(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Projector01()), 1, 2, 1)
        3-qubit circuit with 1 instructions:
        └── ⟨P₀₁(1)⟩ @ q[1,2], z[1]
        <BLANKLINE>
    """

    _name = "Projector01"
    _num_qubits = 2
    _parnames = ()
    _qregsizes = [2]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return se.Matrix([[0, 0, 0, 0], [0, self.a, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"P{to_subscript(0)}{to_subscript(1)}({self.a})"

    def rescale(self, scale):
        return Projector01(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return Projector01(abs(self.a) ** 2)


class Projector10(mc.AbstractOperator):
    r"""Two-qubit operator corresponding to a projection onto :math:`|10\rangle`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & 0\\
            0 & 0 & a & 0\\
            0 & 0 & 0 & 0
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`Projector00`, :class:`Projector01`, :class:`Projector11`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> Projector10()
        P₁₀(1)

        >>> Projector10(0.5)
        P₁₀(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Projector10()), 1, 2, 1)
        3-qubit circuit with 1 instructions:
        └── ⟨P₁₀(1)⟩ @ q[1,2], z[1]
        <BLANKLINE>
    """

    _name = "Projector10"
    _num_qubits = 2
    _parnames = ()
    _qregsizes = [2]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return se.Matrix([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, self.a, 0], [0, 0, 0, 0]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"P{to_subscript(1)}{to_subscript(0)}({self.a})"

    def rescale(self, scale):
        return Projector10(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return Projector10(abs(self.a) ** 2)


class Projector11(mc.AbstractOperator):
    r"""Two-qubit operator corresponding to a projection onto :math:`|11\rangle`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & 0\\
            0 & 0 & 0 & a
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`Projector00`, :class:`Projector01`, :class:`Projector10`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> Projector11()
        P₁₁(1)

        >>> Projector11(0.5)
        P₁₁(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Projector11()), 1, 2, 1)
        3-qubit circuit with 1 instructions:
        └── ⟨P₁₁(1)⟩ @ q[1,2], z[1]
        <BLANKLINE>
    """

    _name = "Projector11"
    _num_qubits = 2
    _parnames = ()
    _qregsizes = [2]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    def _matrix(self):
        return se.Matrix([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, self.a]])

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"P{to_subscript(1)}{to_subscript(1)}({self.a})"

    def rescale(self, scale):
        return Projector11(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale
        return self

    def opsquared(self):
        return Projector11(abs(self.a) ** 2)
    
class ProjectorZ0(mc.AbstractOperator):
    r"""ProjectorZ0()
    Alias for :class:`Projector0`
    """
    def __new__(self):
        return Projector0()


class ProjectorZ1(mc.AbstractOperator):
    r"""ProjectorZ1()
    Alias for :class:`Projector1`
    """
    def __new__(self):
        return Projector1()

__all__ = [
    "Projector0",
    "Projector00",
    "Projector01",
    "Projector1",
    "Projector10",
    "Projector11",
    "ProjectorY1",
    "ProjectorY0",
    "ProjectorX1",
    "ProjectorX0",
    "ProjectorZ1",
    "ProjectorZ0"
]



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
import symengine as se
import numpy as np
import sympy as sp


class Operator(mc.AbstractOperator):
    r"""N-qubit operator specified by a :math:`2^N \times 2^N` matrix.

    .. note::
        Only one and two qubit operators are supported.

    This operator does not have to be unitary.

    See Also:
        :class:`AbstractOperator`, :class:`ExpectationValue`,
        :class:`KrausChannel`

    Parameters:
        matrix (list or np.ndarray): The :math:`2^N \times 2^N` matrix representing the operator.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import Matrix
        >>> op = Operator(Matrix([[1, 2], [3, 4]]))
        >>> op
        1-qubit Operator:
        ├── 1 2
        └── 3 4
        >>> op = Operator(Matrix([[1, 0, 0, 1], [0, 0, 0, 0], [0, 0, 0, 0], [1, 0, 0, 1]]))
        >>> op
        2-qubit Operator:
        ├── 1 0 0 1
        ├── 0 0 0 0
        ├── 0 0 0 0
        └── 1 0 0 1

    Operators can be used for expectation values:

        >>> c = Circuit()
        >>> c.push(ExpectationValue(Operator(Matrix([[0, 1], [0, 0]]))), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨Operator([[0, 1], [0, 0]])⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "Operator"
    _num_qubits = None
    _parnames = ("mat",)
    _num_qregs = 1

    def __init__(self, mat):
        # Check if the input is a valid matrix type (symengine, numpy, or sympy)
        if not isinstance(mat, (se.Matrix, np.ndarray, sp.Matrix)):
            raise TypeError("Must be a symengine, sympy, or numpy Matrix")

        # Convert any numpy or sympy matrices to symengine.Matrix
        mat = self._convert_to_symengine_matrix(mat)
        self.mat = mat

        # Get the matrix dimensions using symengine.Matrix properties
        dim_rows = mat.rows
        dim_cols = mat.cols

        # Check if the matrix is square
        if dim_rows != dim_cols:
            raise ValueError(f"Must be a square matrix, but got {dim_rows}x{dim_cols}")

        # Ensure the dimension is a power of 2
        if not self.is_valid_power_of_2(dim_rows):
            raise ValueError("Dimension of operator has to be 2^n with n>=1")

        # Determine the number of qubits from the matrix dimension
        self.N = int(np.log2(dim_rows))
        if self.N < 1:
            raise ValueError("Cannot define a 0-qubit operator")
        if self.N > 2:
            raise ValueError("Operators larger than 2 qubits are not supported")

        super().__init__()
        self._num_qubits = self.N
        self._qregsizes = [self._num_qubits]

    @staticmethod
    def is_valid_power_of_2(n):
        """Check if a number is a power of 2."""
        return n > 0 and (n & (n - 1)) == 0

    def _convert_to_symengine_matrix(self, matrix):
        """Helper function to convert numpy or sympy matrices to symengine.Matrix."""
        if isinstance(matrix, (np.ndarray, sp.Matrix)):
            return se.Matrix(matrix.tolist())
        return matrix

    def opname(self):
        """Return the operator name."""
        return "Operator"

    def _matrix(self):
        return self.mat

    @property
    def parnames(self):
        """Return the parameter names."""
        return self._parnames

    def iswrapper(self):
        """Placeholder method."""
        pass

    def unwrappedmatrix(self):
        """Return the matrix unwrapped (if needed)."""
        return self.mat

    def __repr__(self):
        """Return a pretty-printed representation."""
        return self.pretty_print()

    def __str__(self):
        """String representation for printing."""
        if self.N < 2:
            return f"{self.opname()}({se.Matrix(self.mat).tolist()})"
        else:
            return f"{self.opname()}(...)"

    def pretty_print(self):
        """Return a detailed string representation of the operator."""
        U = np.array(self.mat.tolist(), dtype=object)
        result = f"{self.N}-qubit Operator:\n"

        rows, cols = U.shape
        for i in range(rows):
            if i < rows - 1:
                result += "├── "
            else:
                result += "└── "
            result += " ".join(map(str, U[i, :]))
            if i < rows - 1:
                result += "\n"
        return result

    def opsquared(self):
        """Return the operator squared (O' * O)."""
        O_dagger = self.mat.T.conjugate()
        squared_matrix = O_dagger * self.mat
        return Operator(squared_matrix)

    def rescale(self, scale):
        """Return a new rescaled Operator by scaling the matrix."""
        scaled_matrix = self.mat * scale
        return Operator(scaled_matrix)

    def rescale_inplace(self, scale):
        """In-place rescaling of the operator matrix."""
        self.mat *= scale
        return self


__all__ = ["Operator"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
from mimiqcircuits.matrices import umatrixpi


class GateID(mc.Gate):
    r"""Single qubit Identity gate.

    **Matrix representation:**

    .. math::
        \operatorname{ID} = \begin{pmatrix}
            1 & 0 \\
            0 & 1
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateID()
        ID
        >>> GateID().matrix()
        [1.0, 0]
        [0, 1.0]
        <BLANKLINE>
        >>> c = Circuit().push(GateID(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── ID @ q[0]
        <BLANKLINE>
        >>> GateID().power(2), GateID().inverse()
        (ID, ID)
        >>> GateID().decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, 0, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "ID"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, n):
        return self

    def isidentity(self):
        return True

    def _matrix(self):
        return umatrixpi(0, 0, 0)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateU(0, 0, 0), q)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.control as mctrl
from mimiqcircuits.operations.gates.standard.cpauli import GateCX
from mimiqcircuits.operations.gates.standard.u import GateU
from mimiqcircuits.operations.gates.standard.phase import GateP


class GateCU(mctrl.Control):
    r"""Two qubit controlled unitary gate.

    **Matrix representation:**

    .. math::
        \operatorname{CU}(\theta, \phi, \lambda, \gamma) = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & e^{i\gamma} \cos\left(\frac{\theta}{2}\right) & -e^{i\gamma} e^{i\lambda}\sin\left(\frac{\theta}{2}\right) \\
            0 & 0 & e^{i\gamma} \mathrm{e}^{i\phi}\sin\left(\frac{\theta}{2}\right) & e^{i\gamma} \mathrm{e}^{i(\phi+\lambda)}\cos\left(\frac{\theta}{2}\right)
        \end{pmatrix}

    Parameters:
        theta (float): Euler angle 1 in radians.
        phi (float): Euler angle 2 in radians.
        lmbda (float): Euler angle 3 in radians.
        gamma (float): Global phase of the CU gate.

    Examples:

        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta, phi, lmbda, gamma = symbols('theta phi lambda gamma')
        >>> GateCU(theta, phi, lmbda, gamma), GateCU(theta, phi, lmbda, gamma).num_controls, GateCU(theta, phi, lmbda, gamma).num_targets, GateCU(theta, phi, lmbda, gamma).num_qubits
        (CU(theta, phi, lambda, gamma), 1, 1, 2)
        >>> GateCU(theta, phi, lmbda, gamma).matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, exp(I*gamma)*cos((1/2)*theta), -exp(I*(gamma + lambda))*sin((1/2)*theta)]
        [0, 0, exp(I*(gamma + phi))*sin((1/2)*theta), exp(I*(gamma + lambda + phi))*cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateCU(theta, phi, lmbda, gamma), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CU(theta, phi, lambda, gamma) @ q[0], q[1]
        <BLANKLINE>
        >>> GateCU(theta, phi, lmbda, gamma).power(2), GateCU(theta, phi, lmbda, gamma).inverse()
        (C(U(theta, phi, lambda, gamma)**2), CU(-theta, -lambda, -phi, -gamma))
        >>> GateCU(theta, phi, lmbda, gamma).decompose()
        2-qubit circuit with 7 instructions:
        ├── P(gamma) @ q[0]
        ├── P((1/2)*(lambda + phi)) @ q[0]
        ├── P((1/2)*(lambda - phi)) @ q[1]
        ├── CX @ q[0], q[1]
        ├── U((-1/2)*theta, 0, (-1/2)*(lambda + phi), 0.0) @ q[1]
        ├── CX @ q[0], q[1]
        └── U((1/2)*theta, phi, 0, 0.0) @ q[1]
        <BLANKLINE>
    """

    def __init__(self, theta, phi, lmbda, gamma):
        super().__init__(1, GateU(theta, phi, lmbda, gamma))

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits
        theta = self.op.theta
        phi = self.op.phi
        lmbda = self.op.lmbda
        gamma = self.op.gamma

        circ.push(GateP(gamma), c)
        circ.push(GateP((lmbda + phi) / 2), c)
        circ.push(GateP((lmbda - phi) / 2), t)
        circ.push(GateCX(), c, t)
        circ.push(GateU(-theta / 2, 0, -(lmbda + phi) / 2), t)
        circ.push(GateCX(), c, t)
        circ.push(GateU(theta / 2, phi, 0), t)

        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.power import Power
from mimiqcircuits.operations.control import Control
from mimiqcircuits.operations.gates.standard.id import GateID


def power_idempotent(op, pow):
    pmod = pow % 2

    if pmod == 0:
        if op.num_qubits < 1:
            raise ValueError("Operation defined on zero qubits.")
        if op.num_qubits == 1:
            return GateID()
        else:
            return GateID().parallel(op.num_qubits)
    elif pmod == 1:
        return op
    elif pow == pmod:
        return Power(op, pmod)
    else:
        return op.power(pmod)


def control_one_defined(n, op, cop, ccop=None, c3op=None):
    if n == 1:
        return cop
    elif n == 2 and ccop is not None:
        return ccop
    elif n == 3 and c3op is not None:
        return c3op
    else:
        return Control(n, op)



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqlink
import tempfile
import json
import os
import shutil
from mimiqcircuits.circuit import Circuit
from mimiqcircuits.qcsresults import QCSResults
from mimiqcircuits.__version__ import __version__
import numpy as np
from time import sleep

# maximum number of samples allowed
MAX_SAMPLES = 2**16

# default value for the number of samples
DEFAULT_SAMPLES = 1000

# minimum and maximum bond dimension allowed
MIN_BONDDIM = 1
MAX_BONDDIM = 2**12

# minimum and maximum entanglement dimension allowed
MIN_ENTDIM = 4
MAX_ENTDIM = 64

# default bond dimension
DEFAULT_BONDDIM = 256

# default entanglement dimension
DEFAULT_ENTDIM = 16

# default time limit
DEFAULT_TIME_LIMIT = 30

# default algorithm
DEFAULT_ALGORITHM = "auto"

RESULTSPB_FILE = "results.pb"

CIRCUIT_FNAME = "circuit"

EXTENSION_PROTO = "pb"
EXTENSION_QASM = "qasm"
EXTENSION_QASM = "stim"

TYPE_PROTO = "proto"
TYPE_QASM = "qasm"
TYPE_STIM = "stim"


class QCSError:
    def __init__(self, error):
        self.error = error

    def __repr__(self):
        return self.error


class MimiqConnection(mimiqlink.MimiqConnection):
    """Represents a connection to the Mimiq Server.

    Inherits from: mimiqlink.MimiqConnection python.
    """

    def __get_timelimit(self):
        """Fetch the maximum time limit for execution from the server."""
        limits = self.user_limits

        if limits and limits.get("enabledMaxTimeout"):
            return limits.get("maxTimeout", DEFAULT_TIME_LIMIT)

        return DEFAULT_TIME_LIMIT

    def execute(
        self,
        circuits,  # Can be a single Circuit object or a list of Circuit objects or QASM file paths
        label="pyapi_v" + __version__,
        algorithm=DEFAULT_ALGORITHM,
        nsamples=DEFAULT_SAMPLES,
        bitstrings=None,
        timelimit=None,
        bonddim=None,
        entdim=None,
        fuse=None,
        reorderqubits=None,
        seed=None,
        qasmincludes=None,
        force=False,
        **kwargs,
    ):
        """
        Execute a circuit or a list of quantum circuits on the Mimiq server.

        Args:
            circuits (Circuit or list of Circuits or str): A single Circuit object, a list of Circuit objects,
                                                           or QASM file paths representing the circuits to be executed.
            label (str): A label for the execution. Defaults to "pyapi_v" + __version__.
            algorithm (str): The algorithm to use. Defaults to "auto".
            nsamples (int): The number of samples to collect. Defaults to DEFAULT_SAMPLES.
            bitstrings (list of str, optional): Specific bitstrings to measure. Defaults to None.
            timelimit (int, optional): The maximum execution time in minutes. Defaults to None.
            bonddim (int, optional): The bond dimension to use. Defaults to None.
            entdim (int, optional): The entanglement dimension to use. Defaults to None.
            fuse (bool, optional): Whether to fuse gates. Defaults to None (let the remote service decide).
            reorderqubits (bool, optional): Whether to reorder qubits. Defaults to None (let the remote service decide).
            seed (int, optional): A seed for random number generation. Defaults to None.
            qasmincludes (list of str, optional): Additional QASM includes. Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            object: A handle to the execution, typically used to retrieve results.

        Raises:
            ValueError: If nsamples exceeds MAX_SAMPLES, bond/entanglement dimensions are out of bounds,
                        or if a circuit contains unevaluated symbolic parameters.
            FileNotFoundError: If a QASM file is not found.
            TypeError: If the circuits argument is not a Circuit object or a valid file path.

        .. note::
                You can also pass a single QASM file path as a string or a list of QASM file paths instead of Circuit objects.
                This allows for executing circuits defined in the OpenQASM format directly.

        Examples:
            ...

            **Connecting to server**

            >>> from mimiqcircuits import *
            >>> conn = MimiqConnection()
            >>> conn.connect()
            Connection:
            ├── url: https://mimiq.qperfect.io/api
            ├── Computing time: 597/10000 minutes
            ├── Executions: 452/10000
            ├── Max time limit per request: 180 minutes
            └── status: open
            <BLANKLINE>
            >>> c = Circuit()
            >>> c.push(GateH(), range(10))
            10-qubit circuit with 10 instructions:
            ├── H @ q[0]
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── H @ q[4]
            ├── H @ q[5]
            ├── H @ q[6]
            ├── H @ q[7]
            ├── H @ q[8]
            └── H @ q[9]
            <BLANKLINE>
            >>> job = conn.execute(c, algorithm="auto")
            >>> res = conn.get_results(job)
            >>> res
            [QCSResults:
            ├── simulator: MIMIQ-StateVector 0.18.0
            ├── timings:
            │    ├── parse time: 7.3677e-05s
            │    ├── apply time: 2.4677e-05s
            │    ├── total time: 0.000306075s
            │    ├── compression time: 8.103e-06s
            │    └── sample time: 0.000141276s
            ├── fidelity estimate: 1
            ├── average multi-qubit gate error estimate: 0
            ├── most sampled:
            │    ├── bs"1100000001" => 7
            │    ├── bs"1010110100" => 5
            │    ├── bs"0010110110" => 4
            │    ├── bs"1001010110" => 4
            │    └── bs"1001100111" => 4
            ├── 1 executions
            ├── 0 amplitudes
            └── 1000 samples]

            **Prepare List of Circuits for execution (Batch-Mode)**

            >>> c1 = Circuit()
            >>> c1.push(Control(2, GateH()), 0, 1, 3)
            4-qubit circuit with 1 instructions:
            └── C₂H @ q[0,1], q[3]
            <BLANKLINE>
            >>> job = conn.execute([c,c1], algorithm="auto")

            List of Results for all Circuits

            >>> res = conn.get_results(job)
            >>> res
            [QCSResults:
            ├── simulator: MIMIQ-StateVector 0.18.0
            ├── timings:
            │    ├── parse time: 0.00011103s
            │    ├── apply time: 2.7497e-05s
            │    ├── total time: 0.000326494s
            │    ├── compression time: 8.657e-06s
            │    └── sample time: 0.000110076s
            ├── fidelity estimate: 1
            ├── average multi-qubit gate error estimate: 0
            ├── most sampled:
            │    ├── bs"1011011100" => 6
            │    ├── bs"0110001101" => 4
            │    ├── bs"0001011001" => 4
            │    ├── bs"1011010010" => 4
            │    └── bs"1000000001" => 4
            ├── 1 executions
            ├── 0 amplitudes
            └── 1000 samples, QCSResults:
            ├── simulator: MIMIQ-StateVector 0.18.0
            ├── timings:
            │    ├── parse time: 0.081798891s
            │    ├── apply time: 0.277081657s
            │    ├── total time: 0.43127283099999997s
            │    ├── amplitudes time: 1.08e-07s
            │    ├── compression time: 0.072055493s
            │    └── sample time: 5.2623e-05s
            ├── fidelity estimate: 1
            ├── average multi-qubit gate error estimate: 0
            ├── most sampled:
            │    └── bs"0000" => 1000
            ├── 1 executions
            ├── 0 amplitudes
            └── 1000 samples]

            Result of the first circuit

            >>> res = conn.get_result(job)
            Warning: Multiple results found. Returning the first one.
            >>> res
            QCSResults:
            ├── simulator: MIMIQ-StateVector 0.18.0
            ├── timings:
            │    ├── parse time: 0.00011103s
            │    ├── apply time: 2.7497e-05s
            │    ├── total time: 0.000326494s
            │    ├── compression time: 8.657e-06s
            │    └── sample time: 0.000110076s
            ├── fidelity estimate: 1
            ├── average multi-qubit gate error estimate: 0
            ├── most sampled:
            │    ├── bs"1011011100" => 6
            │    ├── bs"0110001101" => 4
            │    ├── bs"0001011001" => 4
            │    ├── bs"1011010010" => 4
            │    └── bs"1000000001" => 4
            ├── 1 executions
            ├── 0 amplitudes
            └── 1000 samples

            Input parameters and List of the input Circuits

            >>> circs, parameters = conn.get_inputs(job)
            Downloaded files: ['circuit1.pb', 'circuit2.pb', 'circuits.json', 'request.json']
            >>> circs
            [10-qubit circuit with 10 instructions:
            ├── H @ q[0]
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── H @ q[4]
            ├── H @ q[5]
            ├── H @ q[6]
            ├── H @ q[7]
            ├── H @ q[8]
            └── H @ q[9]
            , 4-qubit circuit with 1 instructions:
            └── C₂H @ q[0,1], q[3]
            ]
            >>> parameters
            {'algorithm': 'auto', 'bitstrings': [], 'samples': 1000, 'seed': 1550300089630762344, 'circuits': [{'file': 'circuit1.pb', 'type': 'proto'}, {'file': 'circuit2.pb', 'type': 'proto'}], 'bondDimension': 256, 'entDimension': 16}

            Input parameters and first input Circuit

            >>> circ, parameters = conn.get_input(job)
            Downloaded files: ['circuit1.pb', 'circuit2.pb', 'circuits.json', 'request.json']
            Warning: Multiple results found. Returning the first one.
            >>> circ
            10-qubit circuit with 10 instructions:
            ├── H @ q[0]
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── H @ q[4]
            ├── H @ q[5]
            ├── H @ q[6]
            ├── H @ q[7]
            ├── H @ q[8]
            └── H @ q[9]
            <BLANKLINE>
            >>> parameters
            {'algorithm': 'auto', 'bitstrings': [], 'samples': 1000, 'seed': 1550300089630762344, 'circuits': [{'file': 'circuit1.pb', 'type': 'proto'}, {'file': 'circuit2.pb', 'type': 'proto'}], 'bondDimension': 256, 'entDimension': 16}


        Connecting Using Credentials
        ----------------------------

        .. code-block:: python

            conn = MimiqConnection(url="https://mimiq.qperfect.io/api")
            conn.connect("Email_address", "Password")

        Saving and Loading Tokens
        -------------------------

        .. code-block:: python

            conn.savetoken("qperfect.json")
            conn.loadtoken("qperfect.json")

        Closing a Connection and Checking Connection Status
        ---------------------------------------------------

        .. code-block:: python

            conn.close()

            conn.isOpen()
        """

        if nsamples > MAX_SAMPLES:
            raise ValueError(f"nsamples must be less than {MAX_SAMPLES}")

        if timelimit is None:
            timelimit = self.__get_timelimit()

        maxtimelimit = self.__get_timelimit()
        if timelimit > maxtimelimit:
            raise ValueError(
                f"Timelimit cannot be set more than {maxtimelimit} minutes."
            )

        if bitstrings is None:
            bitstrings = []
        elif isinstance(circuits, Circuit):
            nq = circuits.num_qubits()
            for b in bitstrings:
                if len(b) != nq:
                    raise ValueError(
                        "The number of qubits in the bitstring is not equal to the number of qubits in the circuit."
                    )

        if seed is None:
            seed = int(np.random.randint(0, np.iinfo(np.int_).max, dtype=np.int_))

        # Set bond and entangling dimensions based on algorithm
        if algorithm == "auto" or algorithm == "mps":
            if bonddim is None:
                bonddim = DEFAULT_BONDDIM
            if entdim is None:
                entdim = DEFAULT_ENTDIM

        if bonddim is not None and (bonddim < MIN_BONDDIM or bonddim > MAX_BONDDIM):
            raise ValueError(f"bonddim must be between {MIN_BONDDIM} and {MAX_BONDDIM}")

        # Check for entdim constraint in specific algorithms
        if algorithm in ["mps", "auto"]:
            actual_entdim = DEFAULT_ENTDIM if entdim is None else entdim
            if actual_entdim < MIN_ENTDIM and not force:
                raise ValueError(
                    f"entdim must be between {MIN_ENTDIM} and {MAX_ENTDIM}"
                )
            elif actual_entdim < MIN_ENTDIM and force:
                print(
                    f"Warning: Running simulation with entdim={actual_entdim}. Results may be misleading."
                )

        with tempfile.TemporaryDirectory() as tmpdir:
            allfiles = []
            circuit_files = []

            # Ensure circuits is a list
            if isinstance(circuits, (Circuit, str)):
                circuits = [circuits]

            if len(circuits) > 1 and algorithm == "auto":
                raise ValueError(
                    "The 'auto' algorithm is not supported in batch mode. Please specify 'mps' or 'statevector' for batch executions."
                )

            if not circuits:
                raise ValueError(
                    "The provided list of circuits is empty. At least one circuit is required."
                )

            for i, c in enumerate(circuits):
                if not isinstance(c, (Circuit, str)):
                    raise TypeError(
                        f"Invalid type at index {i}: expected Circuit or QASM file path, got {type(c).__name__}."
                    )

                if isinstance(c, Circuit) and len(c) == 0:
                    raise ValueError(
                        f"Empty Circuit object at index {i} is not allowed."
                    )

                if isinstance(c, str) and not os.path.isfile(c):
                    raise ValueError(
                        f"Invalid QASM file path at index {i}: {c} does not exist."
                    )

            for i, circuit in enumerate(circuits):
                if isinstance(circuit, Circuit) and circuit.is_symbolic():
                    raise ValueError(
                        "The circuit contains unevaluated symbolic parameters and cannot be processed until all parameters are fully evaluated."
                    )
                if isinstance(circuit, Circuit):
                    circuit_filename = os.path.join(
                        tmpdir, f"{CIRCUIT_FNAME}{i + 1}.{EXTENSION_PROTO}"
                    )
                    circuit.saveproto(circuit_filename)
                    circuit_files.append(
                        {
                            "file": os.path.basename(circuit_filename),
                            "type": TYPE_PROTO,
                        }
                    )
                    allfiles.append(circuit_filename)
                elif isinstance(circuit, str):
                    if not os.path.isfile(circuit):
                        raise FileNotFoundError(f"QASM file {circuit} not found.")
                    circuit_filename = os.path.join(
                        tmpdir, f"{CIRCUIT_FNAME}{i + 1}.{EXTENSION_QASM}"
                    )
                    shutil.copyfile(circuit, circuit_filename)

                    if qasmincludes is None:
                        qasmincludes = []

                    circuit_files.append(
                        {
                            "file": os.path.basename(circuit_filename),
                            "type": TYPE_QASM,
                        }
                    )
                    allfiles.append(circuit_filename)
                else:
                    raise TypeError(
                        "circuits must be Circuit objects or paths to QASM files"
                    )

            jsonbitstrings = ["bs" + o.to01() for o in bitstrings]

            pars = {
                "algorithm": algorithm,
                "bitstrings": jsonbitstrings,
                "samples": nsamples,
                "seed": seed,
                "circuits": circuit_files,
            }

            if bonddim is not None:
                pars["bondDimension"] = bonddim

            if entdim is not None:
                pars["entDimension"] = entdim

            if fuse is not None:
                pars["fuse"] = fuse

            if reorderqubits is not None:
                pars["reorderQubits"] = reorderqubits

            # Add any additional keyword arguments
            pars.update(kwargs)

            # Save the parameters to a JSON file
            pars_filename = os.path.join(tmpdir, "circuits.json")
            with open(pars_filename, "w") as f:
                json.dump(pars, f)

            # Prepare the request
            req = {
                "executor": "Circuits",
                "timelimit": timelimit,
                "apilang": "python",
                "apiversion": __version__,
                "circuitsapiversion": __version__,
            }

            reqfile = os.path.join(tmpdir, "request.json")
            with open(reqfile, "w") as f:
                json.dump(req, f)

            # Make the request to the server
            emutype = "CIRC"

            sleep(0.1)
            return self.request(
                emutype,
                algorithm,
                label,
                timelimit,
                [reqfile, pars_filename] + allfiles,
            )

    def get_results(self, execution, interval=1):
        """Retrieve the results of a completed execution.

        Args:
            execution (str): The execution identifier.
            interval (int): The interval (in seconds) for checking job status (default: 1).

        Returns:
            List[QCSResults]: A list of QCSResults instances.

        Raises:
            RuntimeError: If the remote job encounters an error.
        """
        # Wait for the job to finish
        while not self.isJobDone(execution):
            sleep(interval)

        infos = self.requestInfo(execution)

        if infos["status"] == "ERROR":
            error_message = infos.get("errorMessage", "Remote job errored.")
            raise RuntimeError(f"Remote job errored: {error_message}")
        elif infos["status"] == "CANCELED":
            raise RuntimeError("Remote job canceled.")

        with tempfile.TemporaryDirectory(prefix="mimiq_res_") as tmpdir:
            names = self.downloadResults(execution, destdir=tmpdir)

            results_file_path = os.path.join(tmpdir, "results.json")

            if "results.json" not in names or not os.path.isfile(results_file_path):
                raise RuntimeError(f"No results found in execution {execution}.")

            with open(results_file_path, "r") as f:
                results_list = json.load(f)

            results = []
            for result in results_list:
                if "error" in result:
                    results.append(QCSError(result["error"]))
                else:
                    fname = os.path.join(tmpdir, result["file"])
                    if not os.path.isfile(fname):
                        raise RuntimeError(f"Missing result file {fname}")
                    results.append(QCSResults.loadproto(fname))

        return results

    def get_result(self, execution, **kwargs):
        """Retrieve the first result if multiple are found.

        Args:
            execution (str): The execution identifier.
            **kwargs: Additional keyword arguments for result retrieval.

        Returns:
            QCSResults: The first result found.

        Raises:
            RuntimeWarning: If multiple results are found.
        """
        results = self.get_results(execution, **kwargs)

        if len(results) > 1:
            print("Warning: Multiple results found. Returning the first one.")

        return results[0]

    def get_inputs(self, execution):
        """Retrieve the inputs (circuits and parameters) of the execution.

        Args:
            execution (str): The execution identifier.

        Returns:
            tuple: A tuple containing a list of Circuit objects and parameters (dict).

        Raises:
            RuntimeError: If required files are not found in the inputs.
        """
        with tempfile.TemporaryDirectory(prefix="mimiq_in_") as tmpdir:
            names = self.downloadJobFiles(execution, destdir=tmpdir)

            # Print the downloaded files for debugging purposes
            print(f"Downloaded files: {names}")

            # Get the base names of the downloaded files
            base_names = [os.path.basename(name) for name in names]

            # Check if the circuits.json and request.json files are present
            if "circuits.json" not in base_names or "request.json" not in base_names:
                raise RuntimeError(
                    f"{execution} is not a valid execution for MimiqCircuits: missing necessary files"
                )

            # Load the parameters from the circuits.json file
            circuits_file_path = os.path.join(tmpdir, "circuits.json")
            with open(circuits_file_path, "r") as f:
                parameters = json.load(f)

            circuits = []
            for c in parameters["circuits"]:
                if c["type"] == TYPE_PROTO:
                    circuit = Circuit.loadproto(os.path.join(tmpdir, c["file"]))
                    circuits.append(circuit)
                else:
                    # case of STIM and QASM files
                    circuits.append(os.path.join(tmpdir, c["file"]))

            if len(circuits) == 0:
                raise RuntimeError(
                    "No valid circuit files found. Input parameters not valid."
                )

        return circuits, parameters

    def get_input(self, execution, **kwargs):
        """Retrieve the first circuit and parameters of the execution.

        Args:
            execution (str): The execution identifier.

        Returns:
            tuple: A tuple containing the first Circuit object and parameters (dict).

        Raises:
            RuntimeError: If required files are not found in the inputs.
        """
        circuits, parameters = self.get_inputs(execution, **kwargs)

        if len(circuits) > 1:
            print("Warning: Multiple results found. Returning the first one.")

        return circuits[0], parameters


__all__ = ["MimiqConnection"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.operation import Operation
import mimiqcircuits as mc


class AbstractMeasurement(Operation):

    _name = "AbstractMeasurement"
    def iswrapper(self):
        return False
    
    def __str__(self):
        return self._name
    

class Measure(AbstractMeasurement):
    """Measure operation.

    This operation performs a measurement in the computational basis (Z-basis) and stores the result in a classical register. 

    The measurement projects the quantum state onto either the :math:`|0⟩` or :math:`|1⟩` state, corresponding to the classical bit values of 0 and 1, respectively.

    .. warning::
        `Measure` is non-reversible.

    Examples:

        Adding Measure operation to the Circuit (The qubits (first arg) and the bits (second arg) can be: range, list, tuple, set or int)

        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Measure(),0,0)
        1-qubit circuit with 1 instructions:
        └── M @ q[0], c[0]
        <BLANKLINE>

        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Measure(), range(0,3), range(0,3))
        3-qubit circuit with 3 instructions:
        ├── M @ q[0], c[0]
        ├── M @ q[1], c[1]
        └── M @ q[2], c[2]
        <BLANKLINE>
    """

    _name = "M"
    _num_bits = 1
    _num_qubits = 1
    _num_cregs = 1
    _qregsizes = [1]
    _cregsizes = [1]
    _num_zvars = 0

    def inverse(self):
        raise TypeError("Measure is not inversible")

    def power(self, p):
        raise TypeError("Measure^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled Measure is not defined.")

    def iswrapper(self):
        return False

    def get_operation(self):
        return self


class MeasureX(AbstractMeasurement):
    r"""Single qubit measurement operation in the X-basis.

    This operation projects the quantum state onto the X-basis and stores the result of such
    measurement in a classical register. The measurement is performed by first applying a Hadamard
    gate to rotate the qubit's state into the computational basis, performing a standard Z-basis
    measurement, and then applying another Hadamard gate to return the qubit to its original basis.

    .. warning::
        `MeasureX` is non-reversible.

    See Also:
        :class:`Measure`, :class:`MeasureY`, :class:`MeasureZ`, :class:`Operation`, :class:`Reset`

    Examples:
        >>> from mimiqcircuits import *
        >>> MeasureX()
        MX

        >>> c = Circuit()
        >>> c.push(MeasureX(), 2, 1)
        3-qubit circuit with 1 instructions:
        └── MX @ q[2], c[1]
        <BLANKLINE>

        >>> c.push(MeasureX(), 3, 4)
        4-qubit circuit with 2 instructions:
        ├── MX @ q[2], c[1]
        └── MX @ q[3], c[4]
        <BLANKLINE>
    """

    _name = "MX"
    _num_bits = 1
    _num_qubits = 1
    _num_cregs = 1
    _qregsizes = [1]
    _cregsizes = [1]
    _num_zvars = 0

    def inverse(self):
        raise TypeError("MeasureX is not inversible")

    def power(self, p):
        raise TypeError("MeasureX^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled MeasureX is not defined.")

    def iswrapper(self):
        return False

    def get_operation(self):
        return self

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        b = bits[0]
        circ.push(mc.GateH(), q)
        circ.push(Measure(), q, b)
        circ.push(mc.GateH(), q)
        return circ


class MeasureY(AbstractMeasurement):
    r"""Single qubit measurement operation in the Y-basis.

    This operation projects the quantum state onto the Y-basis and stores the result of such
    measurement in a classical register. The measurement is performed by first applying a :math:`R_y(\pi/2)`
    rotation gate to rotate the qubit's state into the computational basis, performing a standard Z-basis
    measurement, and then applying another :math:`R_y(-\pi/2)` rotation to return the qubit to its original basis.

    .. warning::
        `MeasureY` is non-reversible.

    See Also:
        :class:`Measure`, :class:`MeasureX`, :class:`MeasureZ`, :class:`Operation`, :class:`Reset`

    Examples:
        >>> from mimiqcircuits import *
        >>> MeasureY()
        MY

        >>> c = Circuit()
        >>> c.push(MeasureY(), 2, 1)
        3-qubit circuit with 1 instructions:
        └── MY @ q[2], c[1]
        <BLANKLINE>

        >>> c.push(MeasureY(), 3, 4)
        4-qubit circuit with 2 instructions:
        ├── MY @ q[2], c[1]
        └── MY @ q[3], c[4]
        <BLANKLINE>
    """
    _name = "MY"
    _num_bits = 1
    _num_qubits = 1
    _num_cregs = 1
    _qregsizes = [1]
    _cregsizes = [1]
    _num_zvars = 0

    def inverse(self):
        raise TypeError("MeasureY is not inversible")

    def power(self, p):
        raise TypeError("MeasureY^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled MeasureY is not defined.")

    def iswrapper(self):
        return False

    def get_operation(self):
        return self

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        b = bits[0]
        circ.push(mc.GateHYZ(), q)
        circ.push(Measure(), q, b)
        circ.push(mc.GateHYZ(), q)
        return circ


class MeasureZ(AbstractMeasurement):
    r"""Single qubit measurement operation in the Z-basis.

    This class returns an instance of the :class:`Measure` operation, effectively acting as an alias for Z-basis measurement.

    .. warning::
        `MeasureZ` is non-reversible.

    See Also:
        :class:`Measure`, :class:`MeasureX`, :class:`MeasureY`, :class:`Operation`, :class:`Reset`

    Examples:
        >>> from mimiqcircuits import *
        >>> MeasureZ()
        M

        >>> c = Circuit()
        >>> c.push(MeasureZ(), 0, 0)
        1-qubit circuit with 1 instructions:
        └── M @ q[0], c[0]
        <BLANKLINE>

        >>> c.push(MeasureZ(), 1, 2)
        2-qubit circuit with 2 instructions:
        ├── M @ q[0], c[0]
        └── M @ q[1], c[2]
        <BLANKLINE>
    """

    def __new__(self):
        return Measure()


# export operations
__all__ = ["Measure", "MeasureX", "MeasureY", "MeasureZ"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use thas file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABC, abstractmethod
import copy
import mimiqcircuits as mc
import symengine as se
import sympy as sp
from mimiqcircuits.canvas import _gate_name_padding


class Operation(ABC):
    """
    Abstract base class for quantum operations.
    """

    _name = None

    _num_qubits = None
    _num_qregs = 1
    _qregsizes = None

    _num_bits = None
    _num_cregs = 0
    _cregsizes = None
    _num_zvars = None
    _num_zregs = 0
    _zregsizes = None

    _parnames = ()

    @property
    def num_qubits(self):
        return self._num_qubits

    @num_qubits.setter
    def num_qubits(self, value):
        raise ValueError("Cannot set num_qubits. Read only parameter.")

    @property
    def num_qregs(self):
        return self._num_qregs

    @num_qregs.setter
    def num_qregs(self, value):
        raise ValueError("Cannot set num_qregs. Read only parameter.")

    @property
    def num_bits(self):
        return self._num_bits

    @num_bits.setter
    def num_bits(self, value):
        raise ValueError("Cannot set num_bits. Read only parameter.")

    @property
    def num_zvars(self):
        return self._num_zvars

    @num_zvars.setter
    def num_zvars(self, value):
        raise ValueError("Cannot set num_zvars. Read only parameter.")

    @property
    def num_cregs(self):
        return self._num_cregs

    @num_cregs.setter
    def num_cregs(self, value):
        raise ValueError("Cannot set num_cregs. Read only parameter.")

    @property
    def qregsizes(self):
        return self._qregsizes

    @qregsizes.setter
    def qregsizes(self, value):
        raise ValueError("Cannot set qregsizes. Read only parameter.")

    @property
    def cregsizes(self):
        return self._cregsizes

    @cregsizes.setter
    def cregsizes(self, value):
        raise ValueError("Cannot set cregsizes. Read only parameter.")

    @property
    def zregsizes(self):
        return self._zregsizes

    @zregsizes.setter
    def zregsizes(self, value):
        raise ValueError("Cannot set zregsizes. Read only parameter.")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        raise ValueError("Cannot set name. Read only parameter.")

    @property
    def parnames(self):
        return self._parnames

    @parnames.setter
    def parnames(self, value):
        raise ValueError("Cannot set parnames. Read only parameter.")

    def getparams(self):
        return [getattr(self, pn) for pn in self._parnames]

    def is_symbolic(self):
        return any(
            isinstance(param, (se.Basic, sp.Basic)
                       ) and not param.evalf().is_number
            for param in self.getparams()
        )

    def getparam(self, pn):
        if pn not in self.parnames:
            raise ValueError(f"Parameter {pn} not found.")
        return getattr(self, pn)

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.__dict__ == other.__dict__

    def copy(self):
        """Creates a shallow copy of the operation.
            To create a full copy use deepcopy() instead.

        Returns:
            Operation: A new Operation object containing references to the same attributes as the original circuit
        """
        return copy.copy(self)

    def deepcopy(self):
        """Creates a copy of the object and for all its attributes

        Returns:
            Operation: A new Operation object fully identical the original circuit
        """
        return copy.deepcopy(self)

    @abstractmethod
    def iswrapper(self):
        pass

    def isopalias(self):
        return False

    def numparams(self):
        return len(self.getparams())

    def _decompose(self, circ, qubits, bits, zvars):
        return circ.push(self, *qubits, *bits, *zvars)

    def decompose(self):
        return self._decompose(
            mc.Circuit(),
            range(self.num_qubits),
            range(self.num_bits),
            range(self.num_zvars),
        )

    def evaluate(self, d):
        return self

    def isidentity(self):
        return False

    @classmethod
    def isunitary(self):
        """Check if the class represents a unitary operator.

        By default, this method returns `False` unless explicitly overridden in a subclass.
        """
        return False

    def asciiwidth(self, qubits, bits, zvars):
        # Calculate padding for qubits, bits, and zvars
        namepadding = _gate_name_padding(qubits, bits, zvars)

        # Calculate the width contribution of zvars
        zvar_width = len(",".join(map(str, zvars))) if zvars else 0

        # Assume operation has a string representation (e.g., __str__)
        operation_repr = str(self)

        # Calculate the width: | + (num + space) + name + zvars (if any) + |
        if zvars and not qubits:
            # If there are zvars but no qubits, ensure width includes zvars fully
            return 1 + max(namepadding + len(operation_repr), zvar_width) + 1
        else:
            # Regular case where zvars and qubits are both present or only qubits
            return 1 + namepadding + len(operation_repr) + 1

    def get_operation(self):
        return self


# export operations
__all__ = ["Operation"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from numpy import ndarray
import symengine as se
import sympy as sp
import numpy as np


class GateCustom(mcg.Gate):
    """One or Two qubit Custom gates.

    Examples:
        >>> from mimiqcircuits import Circuit, GateCustom
        >>> import numpy as np
        >>> matrix = np.array([[1, 0, 0, 0],[0, 1, 0, 0],[0, 0, 0, -1j],[0, 0, 1j, 0]])
        >>> c = Circuit()
        >>> c.push(GateCustom(matrix), 0, 1)
        2-qubit circuit with 1 instructions:
        └── Custom(...) @ q[0,1]
        <BLANKLINE>
    """

    _name = "Custom"
    _num_qubits = None
    _qregsizes = None

    def __init__(self, matrix):
        super().__init__()

        if isinstance(matrix, ndarray):
            mat = se.Matrix(matrix.tolist())
        elif isinstance(matrix, se.Matrix):
            mat = matrix
        else:
            raise TypeError(
                f"{type(matrix)} not supported in GateCustom, use symengine.Matrix or numpy.ndarray."
            )

        if mat.rows != mat.cols:
            raise ValueError("Matrix is not square")

        tolerance = 1e-8
        if any(
            isinstance(mat[i, j], (se.Symbol, str))
            for i in range(mat.rows)
            for j in range(mat.cols)
        ):
            pass
        else:
            if not self.is_unitary(mat, tol=tolerance):
                raise ValueError("Matrix is not unitary")

        num_qubits = (mat.rows >> 2) + 1
        if mat.rows != 2**num_qubits:
            raise ValueError("Wrong number of the rows for the matrix")

        self.matrix = mat
        self._num_qubits = num_qubits
        self._qregsizes = [
            num_qubits,
        ]

    @property
    def _matrix(self):
        return self.matrix

    @property
    def num_qubits(self):
        return self._num_qubits

    def inverse(self):
        return GateCustom(self.matrix.inv())

    @staticmethod
    def is_unitary(matrix, tol=1e-8):
        conjugate_transpose = matrix.transpose().conjugate()
        product = matrix * conjugate_transpose
        identity_matrix = se.eye(matrix.rows)

        return all(
            abs(product[i, j] - identity_matrix[i, j]) < tol
            for i in range(matrix.rows)
            for j in range(matrix.cols)
        )

    def __repr__(self):
        return self.pretty_print()

    def __str__(self):
        return f"{self._name}(...)"

    def pretty_print(self):
        if isinstance(self.matrix, se.Matrix):
            U = np.array(self.matrix.tolist(), dtype=object)
        else:
            U = np.array(self.matrix, dtype=object)

        result = f"{self._num_qubits}-qubit GateCustom:\n"
        rows, cols = U.shape
        for i in range(rows):
            if i < rows - 1:
                result += "├── "
            else:
                result += "└── "
            result += " ".join(map(str, U[i, :]))
            if i < rows - 1:
                result += "\n"
        return result

    def evaluate(self, d):
        sympy_matrix = sp.Matrix(self.matrix)
        matrix = sympy_matrix.applyfunc(
            lambda entry: entry.subs(d)
            if not isinstance(entry, (float, int))
            else entry
        )
        evaluated_matrix = se.Matrix(matrix.tolist())
        return GateCustom(evaluated_matrix)


__all__ = ["Custom"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.interactions import GateRZX
from mimiqcircuits.operations.utils import power_idempotent
from mimiqcircuits.operations.gates.standard.pauli import GateX
from symengine import sqrt, I, pi, Matrix


class GateECR(mcg.Gate):
    r"""Two qubit ECR (echo) gate.

    **Matrix representation:**

    .. math::
        \operatorname{ECR} =\begin{pmatrix}
            0 & \frac{1}{\sqrt{2}} & 0 & \frac{i}{\sqrt{2}} \\
            \frac{1}{\sqrt{2}} & 0 & \frac{-i}{\sqrt{2}} & 0 \\
            0 & \frac{i}{\sqrt{2}} & 0 & \frac{1}{\sqrt{2}} \\
            \frac{-i}{\sqrt{2}} & 0 & \frac{1}{\sqrt{2}} & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateECR()
        ECR
        >>> GateECR().matrix()
        [0, 0, 0.707106781186548, 0.0 + 0.707106781186548*I]
        [0, 0, 0.0 + 0.707106781186548*I, 0.707106781186548]
        [0.707106781186548, -0.0 - 0.707106781186548*I, 0, 0]
        [-0.0 - 0.707106781186548*I, 0.707106781186548, 0, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateECR(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── ECR @ q[0,1]
        <BLANKLINE>
        >>> GateECR().power(2), GateECR().inverse()
        (⨷ ² ID, ECR)
        >>> GateECR().decompose()
        2-qubit circuit with 3 instructions:
        ├── RZX((1/4)*pi) @ q[0,1]
        ├── X @ q[0]
        └── RZX((-1/4)*pi) @ q[0,1]
        <BLANKLINE>
    """

    _name = "ECR"

    _num_qubits = 2
    _qregsizes = [2]

    def _matrix(self):
        return Matrix(
            [[0, 0, 1, I], [0, 0, I, 1], [1, -I, 0, 0], [-I, 1, 0, 0]]
        ) / sqrt(2)

    def inverse(self):
        return self

    def _power(self, p):
        return power_idempotent(self, p)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(GateRZX(pi / 4), a, b)
        circ.push(GateX(), a)
        circ.push(GateRZX(-pi / 4), a, b)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.control as mctrl
from mimiqcircuits.operations.gates.standard.cpauli import GateCX
from mimiqcircuits.operations.gates.standard.phase import GateP


class GateCP(mctrl.Control):
    r"""Two qubit Controlled-Phase gate.

    By convention, the first qubit is the control and the second is
    the target

    See Also :func:`GateP`

    **Matrix representation:**

    .. math::
        \operatorname{CP}(\lambda) = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 0 & 0 & e^{i\lambda}
        \end{pmatrix}

    Parameters:
        lambda: Phase angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> lmbda = Symbol('lambda')
        >>> GateCP(lmbda), GateCP(lmbda).num_controls, GateCP(lmbda).num_targets, GateCP(lmbda).num_qubits
        (CP(lambda), 1, 1, 2)
        >>> GateCP(lmbda).matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 1.0, 0]
        [0, 0, 0, exp(I*lambda)]
        <BLANKLINE>
        >>> c = Circuit().push(GateCP(lmbda), 10, 11)
        >>> c
        12-qubit circuit with 1 instructions:
        └── CP(lambda) @ q[10], q[11]
        <BLANKLINE>
        >>> GateCP(lmbda).decompose()
        2-qubit circuit with 5 instructions:
        ├── P((1/2)*lambda) @ q[0]
        ├── CX @ q[0], q[1]
        ├── P((-1/2)*lambda) @ q[1]
        ├── CX @ q[0], q[1]
        └── P((1/2)*lambda) @ q[1]
        <BLANKLINE>
    """

    def __init__(self, lmbda):
        super().__init__(1, GateP(lmbda))

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits
        lmbda2 = self.op.lmbda / 2

        circ.push(GateP(lmbda2), c)
        circ.push(GateCX(), c, t)
        circ.push(GateP(-lmbda2), t)
        circ.push(GateCX(), c, t)
        circ.push(GateP(lmbda2), t)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.gates.standard.s import GateS, GateSDG
from mimiqcircuits.operations.gates.standard.cphase import GateCP
import mimiqcircuits.operations.control as mctrl
from symengine import pi


class GateCS(mctrl.Control):
    r"""Two qubit Controlled-S gate.

    By convention, the first qubit is the control and the second is
    the target

    See Also :func:`GateS`

    **Matrix representation:**:

    .. math::
        \operatorname{CS} =\begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 0 & 0 & i
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCS(), GateCS().num_controls, GateCS().num_targets, GateCS().num_qubits
        (CS, 1, 1, 2)
        >>> GateCS().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 1.0, 0]
        [0, 0, 0, 0.0 + 1.0*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateCS(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CS @ q[0], q[1]
        <BLANKLINE>
        >>> GateCS().power(2), GateCS().inverse()
        (CZ, CS†)
        >>> GateCS().decompose()
        2-qubit circuit with 1 instructions:
        └── CP((1/2)*pi) @ q[0], q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateS())

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(GateCP(pi / 2), a, b)

        return circ


class GateCSDG(mctrl.Control):
    r"""Adjoint of two qubit Controlled-S gate.

    By convention, the first qubit is the control and the second is
    the target

    **Matrix representation:**

    .. math::
        \operatorname{CS}^{\dagger} = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 0 & 0 & -i
            \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCSDG(), GateCSDG().num_controls, GateCSDG().num_targets, GateCSDG().num_qubits
        (CS†, 1, 1, 2)
        >>> GateCSDG().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 1.0, 0]
        [0, 0, 0, 6.12323399573677e-17 - 1.0*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateCSDG(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CS† @ q[0], q[1]
        <BLANKLINE>
        >>> GateCSDG().power(2), GateCSDG().inverse()
        (C(S†**2), CS)
        >>> GateCSDG().decompose()
        2-qubit circuit with 1 instructions:
        └── CP((-1/2)*pi) @ q[0], q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateSDG())

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = range(self.num_qubits)

        circ.push(GateCP(-pi / 2), a, b)

        return circ
